# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Dict, List, Optional

import lightning.pytorch as pl
import numpy as np
import torch
from lightning.pytorch.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
from torch.utils import data
from torch.utils.data import DataLoader, Dataset

from nemo.lightning.pytorch.plugins import MegatronDataSampler

if TYPE_CHECKING:
    from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec


class BERTMockDataModule(pl.LightningDataModule):
    """Mock DataModule for BERT model.
    Args:
        seq_length (int): Sequence length.
        tokenizer (Optional["TokenizerSpec"]): An instance of a TokenizerSpec object.
        micro_batch_size (int): Batch size per GPU.
        global_batch_size (int): Global batch size.
        rampup_batch_size (Optional[List[int]]): Rampup batch size, should be in format of
            [start_global_batch_size, batch_size_increment, ramup_samples].
        num_train_samples (int): Number of training samples.
        num_val_samples (int): Number of validation samples.
        num_test_samples (int): Number of test samples.
        num_workers (int): See ``torch.utils.data.DataLoader`` documentation.
        pin_memory (bool): See ``torch.utils.data.DataLoader`` documentation.
        persistent_workers (bool): See ``torch.utils.data.DataLoader`` documentation.
    """

    def __init__(
        self,
        seq_length: int = 2048,
        tokenizer: Optional["TokenizerSpec"] = None,
        micro_batch_size: int = 4,
        global_batch_size: int = 8,
        rampup_batch_size: Optional[List[int]] = None,
        num_train_samples: int = 10_000,
        num_val_samples: int = 10_000,
        num_test_samples: int = 10_000,
        num_workers: int = 8,
        pin_memory: bool = True,
        persistent_workers: bool = False,
    ):
        super().__init__()
        self.seq_length = seq_length
        self.num_train_samples = num_train_samples
        self.num_val_samples = num_val_samples
        self.num_test_samples = num_test_samples
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.persistent_workers = persistent_workers
        self.global_batch_size = global_batch_size
        self.micro_batch_size = micro_batch_size
        if tokenizer is None:
            from nemo.collections.common.tokenizers.tokenizer_utils import get_nmt_tokenizer

            self.tokenizer = get_nmt_tokenizer("megatron", "BertWordPieceLowerCase")
        else:
            self.tokenizer = tokenizer

        self.data_sampler = MegatronDataSampler(
            seq_len=self.seq_length,
            micro_batch_size=self.micro_batch_size,
            global_batch_size=self.global_batch_size,
            rampup_batch_size=rampup_batch_size,
        )

    def setup(self, stage: str = "") -> None:
        """Setup train/val/test datasets."""
        self._train_ds = _MockBERTDataset(
            self.tokenizer,
            "train",
            self.num_train_samples,
            self.seq_length,
        )
        self._validation_ds = _MockBERTDataset(
            self.tokenizer,
            "valid",
            self.num_val_samples,
            self.seq_length,
        )
        self._test_ds = _MockBERTDataset(
            self.tokenizer,
            "test",
            self.num_test_samples,
            self.seq_length,
        )

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """Create train dataloader."""
        if not hasattr(self, "_train_ds"):
            self.setup()
        return self._create_dataloader(self._train_ds)

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """Create validation dataloader."""
        if not hasattr(self, "_validation_ds"):
            self.setup()
        return self._create_dataloader(self._validation_ds)

    def test_dataloader(self) -> EVAL_DATALOADERS:
        """Create test dataloader."""
        if not hasattr(self, "_test_ds"):
            self.setup()
        return self._create_dataloader(self._test_ds)

    def _create_dataloader(self, dataset, **kwargs) -> DataLoader:
        return DataLoader(
            dataset,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            persistent_workers=self.persistent_workers,
            collate_fn=dataset.collate_fn,
            **kwargs,
        )

    def reconfigure_limit_batches(self):
        """
        Reconfigure trainer.limit_train_batches and trainer.limit_val_batches in terms of num of microbatches.
        """
        from nemo.collections.llm.gpt.data.utils import _reconfigure_limit_batches

        # Override limit_train_batches in terms of num of microbatches
        self.trainer.limit_train_batches = _reconfigure_limit_batches(self.trainer.limit_train_batches, self._train_ds)
        # Override limit_val_batches to be a multiple of num microbatches to prevent val_step from exiting
        #   in between a step
        self.trainer.limit_val_batches = _reconfigure_limit_batches(
            self.trainer.limit_val_batches, self._validation_ds
        )

        try:
            from megatron.core.num_microbatches_calculator import get_num_microbatches

        except (ImportError, ModuleNotFoundError):
            from apex.transformer.pipeline_parallel.utils import get_num_microbatches

        # Override num sanity steps to be a multiple of num of microbatches
        self.trainer.num_sanity_val_steps *= get_num_microbatches()


class _MockBERTDataset(Dataset):
    def __init__(
        self,
        tokenizer: "TokenizerSpec",
        name: str,
        num_samples: int,
        seq_length: int,
        seed: int = 420,
    ) -> None:
        super().__init__()
        self.name = name
        self.seq_length = seq_length
        self.vocab_size = tokenizer.vocab_size
        self.length = num_samples
        self.seed = seed
        self.loss_mask = torch.ones(self.seq_length, dtype=torch.float)
        self.position_ids = torch.arange(self.seq_length, dtype=torch.int64)

    def __len__(self) -> int:
        return self.length

    def _get_text(self, idx: int) -> np.ndarray:
        np_gen = np.random.default_rng(seed=(self.seed + idx))
        return np_gen.integers(self.vocab_size, size=[self.seq_length], dtype=np.int64)

    def __getitem__(self, idx) -> Dict[str, torch.Tensor]:
        # Generate data of the expected size and datatype (based on GPTDataset).
        np_gen = np.random.default_rng(seed=(self.seed + idx))
        tokens = torch.from_numpy(np_gen.integers(self.vocab_size, size=[self.seq_length], dtype=np.int64))
        labels = torch.from_numpy(np_gen.integers(self.vocab_size, size=[self.seq_length], dtype=np.int64))
        assignments = torch.zeros(self.seq_length, dtype=torch.int64)
        is_next_random = np_gen.random() < 0.5
        mask_pads = torch.from_numpy(np.ones([self.seq_length], dtype=np.int64))
        truncated = np_gen.random() < 0.5

        batch = {
            "text": tokens,
            "types": assignments,
            "labels": labels,
            "is_random": int(is_next_random),
            "padding_mask": mask_pads,
            "loss_mask": self.loss_mask,
            "truncated": int(truncated),
        }

        return batch

    def _collate_fn(self, batch):
        """
        A default implementation of a collation function.
        Users should override this method to define custom data loaders.
        """
        return data.dataloader.default_collate(batch)

    def collate_fn(self, batch):
        """Method that user pass as functor to DataLoader.

        The method optionally performs neural type checking and add types to the outputs.

        Please note, subclasses of Dataset should not implement `input_types`.

        # Usage:
        dataloader = torch.utils.data.DataLoader(
                ....,
                collate_fn=dataset.collate_fn,
                ....
        )

        Returns
        -------
            Collated batch, with or without types.
        """
        return self._collate_fn(batch)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re

import pytest
import torch

from nemo.collections.asr.inference.utils.pipeline_utils import (
    check_existance_of_required_attributes,
    drop_trailing_features,
    get_leading_punctuation_regex_pattern,
    get_repeated_punctuation_regex_pattern,
)


class TestPipelineUtils:

    @pytest.mark.unit
    def test_drop_trailing_features(self):
        x = torch.randn(10, 10, 20)
        expected_feature_buffer_len = 15
        x_dropped = drop_trailing_features(x, expected_feature_buffer_len)
        assert x_dropped.shape == (10, 10, 15)
        assert x_dropped.allclose(x[:, :, :15])

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "text, expected_text",
        [
            ("", ""),
            (" ", " "),
            ("simple text", "simple text"),
            ("just a 2nd . Yeah, I hope", "just a 2nd. Yeah, I hope"),
            ("Hello , world ! How are you ?", "Hello, world! How are you?"),
            ("The quick, brown fox jumps ? over the lazy ! dog.", "The quick, brown fox jumps? over the lazy! dog."),
        ],
    )
    def test_remove_leading_punctuation_spaces(self, text, expected_text):
        puncts = {"!", "?", ".", ","}
        pattern = get_leading_punctuation_regex_pattern(puncts)
        assert re.sub(pattern, r'\1', text) == expected_text

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "text, expected_text",
        [
            ("", ""),
            (" ", " "),
            ("simple text", "simple text"),
            ("Hello, world!! How are you???", "Hello, world! How are you?"),
            ("The quick,, brown fox jumps? over the lazy! dog..", "The quick, brown fox jumps? over the lazy! dog."),
        ],
    )
    def test_remove_repeated_punctuation(self, text, expected_text):
        puncts = {"!", "?", ".", ","}
        pattern = get_repeated_punctuation_regex_pattern(puncts)
        assert re.sub(pattern, r'\1', text) == expected_text

    @pytest.mark.unit
    def test_check_existance_of_required_attributes(self):
        class TestClass:
            pass

        with pytest.raises(ValueError):
            check_existance_of_required_attributes(TestClass, ['test_attr'])



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cache

import torch
from lightning.pytorch.callbacks.callback import Callback

try:
    from megatron.core.utils import check_param_hashes_across_dp_replicas

    HAVE_MEGATRON_CORE = True

except (ImportError, ModuleNotFoundError):

    HAVE_MEGATRON_CORE = False

from nemo.lightning import io
from nemo.utils import logging


@cache
def pl_has_dist_opt_with_ovelap(trainer):
    optim_config = getattr(getattr(trainer.strategy.model, 'optim', None), 'config', None)
    if not getattr(optim_config, 'use_distributed_optimizer', False):
        return False
    if not getattr(optim_config, 'overlap_param_gather', False):
        return False
    return True


def pl_check_param_hashes_across_dp_replicas(trainer):
    if pl_has_dist_opt_with_ovelap(trainer):
        for opt in self.optimizers:
            opt.disable_pre_hook()
    import megatron.core.parallel_state as mp

    res = check_param_hashes_across_dp_replicas([trainer.strategy.model])
    torch.distributed.barrier()

    all_res = [False for _ in range(mp.get_data_parallel_world_size())]

    torch.distributed.all_gather_object(all_res, res, group=mp.get_data_parallel_group_gloo())

    if pl_has_dist_opt_with_ovelap(trainer):
        for opt in self.optimizers:
            opt.enable_pre_hook()
    return all(all_res)


class DdpParityChecker(Callback, io.IOMixin):
    """
    This callback enables weight parity checkping across DDP replicas with Mcore models.

    User can specify their desired interval for weights to be checked via the `interval` parameter.

    Args:
        dir (Optional[str]): Directory to store the memory profile dump

    Example:
        >>> callback = DdpParityChecker(interval=10)
        >>> trainer = Trainer(callbacks=[callback])
    """

    def __init__(self, interval: int = 0):
        """
        interval (int): How frequently to check DDP weights for errors. Default to 0 (off).
        """
        assert interval > 0, "Expected interval to be > 0. A zero interval makes DdpParityChecker a no-op."
        self.interval = interval
        self.step = 0

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx, unused=0) -> None:
        if self.step == self.interval - 1:
            if pl_check_param_hashes_across_dp_replicas(trainer):
                logging.info(f"DDP Param parity check passed for batch-id= {batch_idx}")
            else:
                trainer.should_stop = True
                trainer.limit_val_batches = 0
                logging.info(f"DDP Param parity check FAILED for batch-id= {batch_idx}")
        self.step = (self.step + 1) % self.interval

    def on_train_end(self, trainer, pl_module) -> None:
        pl_check_param_hashes_across_dp_replicas(trainer)
        logging.info("DDP Param parity check passed at end of training.")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Tuple

import torch


def _pad_attention_masks(
    masks: torch.Tensor,
    num_chunks: torch.Tensor,
    total_length: int,
    max_chunks: int,
    device: torch.device,
    dtype=torch.bfloat16,
) -> torch.Tensor:
    """
    Pads the provided masks to a uniform shape for batching.

    Args:
        masks (torch.Tensor): List of tensors containing attention masks for each batch.
        num_chunks (torch.Tensor): Tensor containing the number of chunks for each mask.
        total_length (int): Total sequence length for padding.
        max_chunks (int): Maximum number of chunks to pad each mask to.
        device (torch.device): Device to place the output tensor on.
        dtype (torch.dtype): Data type for the output tensor. Default is `torch.bfloat16`.

    Returns:
        torch.Tensor: A padded tensor of shape [B, total_length, max_num_media, max_chunks]
        where `B` is the batch size.
    """
    mask_value = 1.0
    batch_size = len(masks)
    max_num_media = max([len(m) for m in masks])

    padded_masks = torch.full(
        (batch_size, total_length, max_num_media, max_chunks),
        mask_value,
        dtype=dtype,
        device=device,
    )

    for idx, (mask_group, chunks) in enumerate(zip(masks, num_chunks)):
        for media_idx, (mask, chunk_count) in enumerate(zip(mask_group, chunks)):
            if len(mask) == 2:
                mask[1] = min(mask[1], total_length)
                if mask[1] == -1:
                    mask[1] = total_length
                padded_masks[idx, mask[0] : mask[1], media_idx, :chunk_count].fill_(0.0)

    return padded_masks


def _get_full_row_masked_out_mask(
    attention_bias: torch.Tensor,
    mask_value: float,
):
    """
    Determines whether each row in the attention bias tensor contains masked values.

    Args:
        attention_bias (torch.Tensor): A 4D tensor of shape [B, H, S1, S2], where:
            - B: Batch size.
            - H: Number of attention heads.
            - S1: Length of the first sequence.
            - S2: Length of the second sequence.
        mask_value (float): The value used to represent masked positions in `attention_bias`.

    Returns:
        torch.Tensor: A 4D tensor of shape [B, H, S1, 1], containing boolean values (as a tensor)
        indicating if each row in the last dimension is fully masked (0 if fully masked, 1 otherwise).
    """
    return (attention_bias != mask_value).any(dim=-1).type_as(attention_bias)[..., None]


def _generate_cross_attention_mask(
    text_token_count: int,
    text_device: torch.device,
    text_dtype: torch.dtype,
    vision_tokens: torch.Tensor,
    cross_attention_masks: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Generates a cross-attention mask for aligning text and vision tokens.

    Args:
        text_token_count (int): Number of tokens in the text sequence.
        text_device (torch.device): Device to place the output tensor on.
        text_dtype (torch.dtype): Data type for the output tensor.
        vision_tokens (torch.Tensor): Vision tokens tensor of shape [B, I, T, D] where:
            - B: Batch size.
            - I: Number of images.
            - T: Number of image tokens per image.
            - D: Dimension of each image token.
        cross_attention_masks (torch.Tensor): Cross attention masks of shape [B, N, I, C], where:
            - B: Batch size.
            - N: Number of text tokens.
            - I: Number of images.
            - C: Number of chunks.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: A tuple containing:
            - The adjusted cross-attention masks of shape [B, 1, N, I * T].
            - The full row mask status tensor of shape [B, 1, N, 1].
    """
    assert vision_tokens is not None, "Vision tokens must be provided"
    vision_token_length = vision_tokens.shape[3]
    assert (
        vision_tokens.shape[1] == cross_attention_masks.shape[2]
    ), f"Mismatch in number of images given and number of masks provided: {vision_tokens.shape} vs {cross_attention_masks.shape}"
    assert (
        vision_tokens.shape[2] == cross_attention_masks.shape[3]
    ), f"Mismatch between vision tokens and cross-attention masks: {vision_tokens.shape} vs {cross_attention_masks.shape}"
    assert (
        text_token_count == cross_attention_masks.shape[1]
    ), f"Text sequence length {text_token_count} does not match cross-attention mask length {cross_attention_masks.shape[1]}"

    batch_size, _, num_images, num_chunks = cross_attention_masks.shape
    cross_attention_masks = cross_attention_masks.view(batch_size, text_token_count, -1).unsqueeze(1)

    full_row_mask_status = _get_full_row_masked_out_mask(cross_attention_masks, mask_value=1.0)
    cross_attention_masks = cross_attention_masks.repeat_interleave(vision_token_length, dim=3)
    cross_attention_masks *= full_row_mask_status

    return (
        cross_attention_masks.to(device=text_device, dtype=text_dtype),
        full_row_mask_status.to(device=text_device, dtype=text_dtype),
    )


def create_vision_mask_tensor(tokens: torch.Tensor, vision_token_id: int = 128256) -> torch.Tensor:
    """
    Create a vision mask from a tensor of tokens and a vision token ID.

    Args:
        tokens (torch.Tensor): A 1D tensor of token IDs.
        vision_token_id (int): The ID of the vision token.

    Returns:
        torch.Tensor: A tensor containing vision masks in the format [start, end].
    """
    # Get the locations of the vision tokens
    vision_token_locations = (tokens == vision_token_id).nonzero(as_tuple=False).squeeze()

    # If no vision token found, return an empty tensor
    if vision_token_locations.numel() == 0:
        return torch.empty(1, 2, dtype=torch.long)

    vision_masks = []

    # Handle case with only one vision token
    if vision_token_locations.numel() == 1:
        vision_masks.append([vision_token_locations.item(), len(tokens)])
    else:
        # Multiple vision tokens, pairwise masks
        for i in range(len(vision_token_locations) - 1):
            vision_masks.append([vision_token_locations[i].item(), vision_token_locations[i + 1].item()])
        # Last vision token attends to all subsequent text
        vision_masks.append([vision_token_locations[-1].item(), len(tokens)])

    # Handle consecutive vision tokens
    last_mask_end = vision_masks[-1][1]
    for vision_mask in reversed(vision_masks):
        if vision_mask[0] == vision_mask[1] - 1:
            vision_mask[1] = last_mask_end
        last_mask_end = vision_mask[1]

    return torch.tensor(vision_masks, dtype=torch.long)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect
from typing import Callable, Optional
from lightning.pytorch.callbacks import LambdaCallback


class ModelCallback(LambdaCallback):
    """
    A callback that extends LambdaCallback to intelligently handle function parameters.
    Functions can take either (trainer, pl_module), just (pl_module), or just (trainer).

    Supported parameter names:
    - trainer, pl_trainer
    - model, pl_model, pl_module, module

    Example:
        >>> # Using with torch.compile
        >>> callback = ModelCallback(on_train_start=torch.compile)
        >>>
        >>> # Using with thunder_compile
        >>> callback = ModelCallback(on_train_start=thunder_compile)
        >>>
        >>> # Mix different callbacks
        >>> callback = ModelCallback(
        ...     on_train_start=lambda model: torch.compile(model),
        ...     on_fit_start=lambda trainer, model: print(f"Starting fit with {model}")
        ... )
    """

    TRAINER_PARAMS = {'trainer', 'pl_trainer'}
    MODEL_PARAMS = {'model', 'pl_model', 'pl_module', 'module'}

    def __init__(
        self,
        setup: Optional[Callable] = None,
        teardown: Optional[Callable] = None,
        on_fit_start: Optional[Callable] = None,
        on_fit_end: Optional[Callable] = None,
        on_sanity_check_start: Optional[Callable] = None,
        on_sanity_check_end: Optional[Callable] = None,
        on_train_batch_start: Optional[Callable] = None,
        on_train_batch_end: Optional[Callable] = None,
        on_train_epoch_start: Optional[Callable] = None,
        on_train_epoch_end: Optional[Callable] = None,
        on_validation_epoch_start: Optional[Callable] = None,
        on_validation_epoch_end: Optional[Callable] = None,
        on_test_epoch_start: Optional[Callable] = None,
        on_test_epoch_end: Optional[Callable] = None,
        on_validation_batch_start: Optional[Callable] = None,
        on_validation_batch_end: Optional[Callable] = None,
        on_test_batch_start: Optional[Callable] = None,
        on_test_batch_end: Optional[Callable] = None,
        on_train_start: Optional[Callable] = None,
        on_train_end: Optional[Callable] = None,
        on_validation_start: Optional[Callable] = None,
        on_validation_end: Optional[Callable] = None,
        on_test_start: Optional[Callable] = None,
        on_test_end: Optional[Callable] = None,
        on_exception: Optional[Callable] = None,
        on_save_checkpoint: Optional[Callable] = None,
        on_load_checkpoint: Optional[Callable] = None,
        on_before_backward: Optional[Callable] = None,
        on_after_backward: Optional[Callable] = None,
        on_before_optimizer_step: Optional[Callable] = None,
        on_before_zero_grad: Optional[Callable] = None,
        on_predict_start: Optional[Callable] = None,
        on_predict_end: Optional[Callable] = None,
        on_predict_batch_start: Optional[Callable] = None,
        on_predict_batch_end: Optional[Callable] = None,
        on_predict_epoch_start: Optional[Callable] = None,
        on_predict_epoch_end: Optional[Callable] = None,
    ):
        # Create a dictionary of non-None callbacks
        callbacks = {
            name: self._wrap_func(func)
            for name, func in locals().items()
            if name != 'self' and name != '__class__' and func is not None
        }

        super().__init__(**callbacks)

    def _get_param_type(self, param_name: str) -> Optional[str]:
        """Determine if a parameter name refers to trainer or model."""
        param_name = param_name.lower()
        if param_name in self.TRAINER_PARAMS:
            return 'trainer'
        if param_name in self.MODEL_PARAMS:
            return 'model'
        return None

    def _wrap_func(self, func: Callable) -> Callable:
        """Wraps a function to handle parameter inspection and passing."""
        sig = inspect.signature(func)
        params = sig.parameters

        def wrapped(trainer, pl_module, *args, **kwargs):
            call_args = {}

            for param_name, param in params.items():
                param_type = self._get_param_type(param_name)

                if param_type == 'trainer':
                    call_args[param_name] = trainer
                elif param_type == 'model':
                    call_args[param_name] = pl_module
                else:
                    # If parameter name is not recognized, use position to determine
                    if len(params) == 1:
                        call_args[param_name] = pl_module
                    elif len(params) == 2:
                        if len(call_args) == 0:
                            call_args[param_name] = trainer
                        else:
                            call_args[param_name] = pl_module
                    else:
                        raise ValueError(
                            f"Unable to determine parameter mapping for '{param_name}'. "
                            f"Please use recognized parameter names: "
                            f"trainer/pl_trainer for trainer, "
                            f"model/pl_model/pl_module/module for model."
                        )

            try:
                return func(**call_args)
            except TypeError as e:
                raise TypeError(
                    f"Failed to call callback function {func.__name__ if hasattr(func, '__name__') else func}. "
                    f"Attempted to pass arguments: {call_args.keys()}. Error: {str(e)}"
                ) from e

        return wrapped



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
from functools import cached_property, lru_cache
from pathlib import Path
from typing import Optional

import pytest
import torch
from omegaconf import DictConfig

from nemo.collections.asr.models import ASRModel
from nemo.collections.asr.modules import RNNTDecoder, RNNTJoint
from nemo.collections.asr.parts.context_biasing import BoostingTreeModelConfig, GPUBoostingTreeModel
from nemo.collections.asr.parts.mixins import mixins
from nemo.collections.asr.parts.submodules import rnnt_beam_decoding
from nemo.collections.asr.parts.submodules import rnnt_greedy_decoding as greedy_decode
from nemo.collections.asr.parts.submodules import tdt_beam_decoding
from nemo.collections.asr.parts.submodules.ngram_lm import NGramGPULanguageModel
from nemo.collections.asr.parts.submodules.rnnt_decoding import RNNTBPEDecoding, RNNTDecoding, RNNTDecodingConfig
from nemo.collections.asr.parts.utils import rnnt_utils
from nemo.core.utils import numba_utils
from nemo.core.utils.numba_utils import __NUMBA_MINIMUM_VERSION__
from tests.collections.asr.decoding.test_timestamps import BaseTimestampsTest

NUMBA_RNNT_LOSS_AVAILABLE = numba_utils.numba_cpu_is_supported(
    __NUMBA_MINIMUM_VERSION__
) or numba_utils.numba_cuda_is_supported(__NUMBA_MINIMUM_VERSION__)


def char_vocabulary():
    return [' ', 'a', 'b', 'c', 'd', 'e', 'f', '.']


@pytest.fixture()
@lru_cache(maxsize=8)
def tmp_tokenizer(test_data_dir):
    cfg = DictConfig({'dir': os.path.join(test_data_dir, "asr", "tokenizers", "an4_wpe_128"), 'type': 'wpe'})

    class _TmpASRBPE(mixins.ASRBPEMixin):
        def register_artifact(self, _, vocab_path):
            return vocab_path

    asrbpe = _TmpASRBPE()
    asrbpe._setup_tokenizer(cfg)
    return asrbpe.tokenizer


@lru_cache(maxsize=2)
def get_rnnt_decoder(vocab_size, decoder_output_size=4):
    prednet_cfg = {'pred_hidden': decoder_output_size, 'pred_rnn_layers': 1}
    torch.manual_seed(0)
    decoder = RNNTDecoder(prednet=prednet_cfg, vocab_size=vocab_size)
    decoder.freeze()
    return decoder


@lru_cache(maxsize=2)
def get_rnnt_joint(vocab_size, vocabulary=None, encoder_output_size=4, decoder_output_size=4, joint_output_shape=4):
    jointnet_cfg = {
        'encoder_hidden': encoder_output_size,
        'pred_hidden': decoder_output_size,
        'joint_hidden': joint_output_shape,
        'activation': 'relu',
    }
    torch.manual_seed(0)
    joint = RNNTJoint(jointnet_cfg, vocab_size, vocabulary=vocabulary)
    joint.freeze()
    return joint


@lru_cache(maxsize=1)
def get_model_encoder_output(data_dir, model_name):
    # Import inside function to avoid issues with dependencies
    import librosa

    audio_filepath = os.path.join(data_dir, 'asr', 'test', 'an4', 'wav', 'cen3-fjlp-b.wav')

    with torch.no_grad():
        model = ASRModel.from_pretrained(model_name, map_location='cpu')  # type: ASRModel
        model.preprocessor.featurizer.dither = 0.0
        model.preprocessor.featurizer.pad_to = 0
        model.eval()

        audio, sr = librosa.load(path=audio_filepath, sr=16000, mono=True)

        input_signal = torch.tensor(audio, dtype=torch.float32).unsqueeze(0)
        input_signal_length = torch.tensor([len(audio)], dtype=torch.int32)

        encoded, encoded_len = model(input_signal=input_signal, input_signal_length=input_signal_length)

    return model, encoded, encoded_len


def decode_text_from_greedy_hypotheses(hyps, decoding):
    decoded_hyps = decoding.decode_hypothesis(hyps)  # type: List[str]

    return decoded_hyps


def decode_text_from_nbest_hypotheses(hyps, decoding):
    hypotheses = []
    all_hypotheses = []

    for nbest_hyp in hyps:  # type: rnnt_utils.NBestHypotheses
        n_hyps = nbest_hyp.n_best_hypotheses  # Extract all hypotheses for this sample
        decoded_hyps = decoding.decode_hypothesis(n_hyps)  # type: List[str]

        hypotheses.append(decoded_hyps[0])  # best hypothesis
        all_hypotheses.append(decoded_hyps)

    return hypotheses, all_hypotheses


def check_beam_decoding(test_data_dir, beam_config):
    beam_size = beam_config.pop("beam_size", 1)
    model, encoded, encoded_len = get_model_encoder_output(test_data_dir, 'nvidia/parakeet-tdt_ctc-110m')

    model_config = model.to_config_dict()
    durations = list(model_config["model_defaults"]["tdt_durations"])

    beam = tdt_beam_decoding.BeamTDTInfer(
        model.decoder,
        model.joint,
        beam_size=beam_size,
        return_best_hypothesis=False,
        durations=durations,
        **beam_config,
    )

    enc_out = encoded
    enc_len = encoded_len

    with torch.no_grad():
        hyps: rnnt_utils.Hypothesis = beam(encoder_output=enc_out, encoded_lengths=enc_len)[0]
        _, all_hyps = decode_text_from_nbest_hypotheses(hyps, model.decoding)
        all_hyps = all_hyps[0]

        print("Beam search algorithm :", beam_config['search_type'])
        for idx, hyp_ in enumerate(all_hyps):
            print("Hyp index", idx + 1, "text :", hyp_.text)

            assert len(hyp_.timestamp) > 0
            print("Timesteps", hyp_.timestamp)
            print()


def check_tdt_greedy_decoding(
    test_data_dir,
    use_cuda_graph_decoder: bool,
    lm_path: Optional[str | Path] = None,
    boosting_tree: Optional[BoostingTreeModelConfig] = None,
    enable_per_stream_biasing: bool = False,
):
    model, encoded, encoded_len = get_model_encoder_output(test_data_dir, 'nvidia/parakeet-tdt_ctc-110m')

    model_config = model.to_config_dict()

    fusion_models, fusion_models_alpha = None, None
    if lm_path or boosting_tree:
        fusion_models = []
        fusion_models_alpha = []

    if lm_path:
        fusion_models.append(NGramGPULanguageModel.from_file(lm_path=lm_path, vocab_size=model.decoder.blank_idx))
        fusion_models_alpha.append(0.5)
    if boosting_tree:
        fusion_models.append(GPUBoostingTreeModel.from_config(boosting_tree, tokenizer=model.tokenizer))
        fusion_models_alpha.append(0.5)

    decoding_algo = greedy_decode.GreedyBatchedTDTInfer(
        model.decoder,
        model.joint,
        blank_index=model.decoder.blank_idx,
        durations=list(model_config["model_defaults"]["tdt_durations"]),
        max_symbols_per_step=10,
        preserve_alignments=False,
        preserve_frame_confidence=False,
        use_cuda_graph_decoder=use_cuda_graph_decoder,
        fusion_models=fusion_models,
        fusion_models_alpha=fusion_models_alpha,
        enable_per_stream_biasing=enable_per_stream_biasing,
    )

    enc_out = encoded
    enc_len = encoded_len

    with torch.no_grad():
        hyps: rnnt_utils.Hypothesis = decoding_algo(encoder_output=enc_out, encoded_lengths=enc_len)[0]
        all_hyps = decode_text_from_greedy_hypotheses(hyps, model.decoding)

        print("Decoding result")
        for idx, hyp_ in enumerate(all_hyps):
            print(f"Hyp index {idx + 1} | text : {hyp_.text}")
            assert len(hyp_.timestamp) > 0
            print("Timesteps", hyp_.timestamp)
            print()


class TestRNNTDecoding:
    @pytest.mark.unit
    def test_constructor(self):
        cfg = RNNTDecodingConfig()
        vocab = char_vocabulary()
        decoder = get_rnnt_decoder(vocab_size=len(vocab))
        joint = get_rnnt_joint(vocab_size=len(vocab))
        decoding = RNNTDecoding(decoding_cfg=cfg, decoder=decoder, joint=joint, vocabulary=vocab)
        assert decoding is not None

    @pytest.mark.unit
    def test_constructor_subword(self, tmp_tokenizer):
        cfg = RNNTDecodingConfig()
        vocab = tmp_tokenizer.vocab
        decoder = get_rnnt_decoder(vocab_size=len(vocab))
        joint = get_rnnt_joint(vocab_size=len(vocab))
        decoding = RNNTBPEDecoding(decoding_cfg=cfg, decoder=decoder, joint=joint, tokenizer=tmp_tokenizer)
        assert decoding is not None

    @pytest.mark.skipif(
        not NUMBA_RNNT_LOSS_AVAILABLE,
        reason='RNNTLoss has not been compiled with appropriate numba version.',
    )
    @pytest.mark.with_downloads
    @pytest.mark.unit
    def test_greedy_decoding_preserve_alignments(self, test_data_dir):
        model, encoded, encoded_len = get_model_encoder_output(test_data_dir, 'stt_en_conformer_transducer_small')

        beam = greedy_decode.GreedyRNNTInfer(
            model.decoder,
            model.joint,
            blank_index=model.joint.num_classes_with_blank - 1,
            max_symbols_per_step=5,
            preserve_alignments=True,
        )

        enc_out = encoded
        enc_len = encoded_len

        with torch.no_grad():
            hyps = beam(encoder_output=enc_out, encoded_lengths=enc_len)[0]  # type: rnnt_utils.Hypothesis
            hyp = decode_text_from_greedy_hypotheses(hyps, model.decoding)
            hyp = hyp[0]

            assert hyp.alignments is not None

            # Use the following commented print statements to check
            # the alignment of other algorithms compared to the default
            print("Text", hyp.text)
            for t in range(len(hyp.alignments)):
                t_u = []
                for u in range(len(hyp.alignments[t])):
                    logp, label = hyp.alignments[t][u]
                    assert torch.is_tensor(logp)
                    assert torch.is_tensor(label)

                    t_u.append(int(label))

                print(f"Tokens at timestamp {t} = {t_u}")
            print()

    @pytest.mark.skipif(
        not NUMBA_RNNT_LOSS_AVAILABLE,
        reason='RNNTLoss has not been compiled with appropriate numba version.',
    )
    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize("loop_labels", [True, False])
    def test_batched_greedy_decoding_preserve_alignments(self, test_data_dir, loop_labels: bool):
        """Test batched greedy decoding using non-batched decoding as a reference"""
        model, encoded, encoded_len = get_model_encoder_output(test_data_dir, 'stt_en_conformer_transducer_small')

        search_algo = greedy_decode.GreedyBatchedRNNTInfer(
            model.decoder,
            model.joint,
            blank_index=model.joint.num_classes_with_blank - 1,
            max_symbols_per_step=5,
            preserve_alignments=True,
            loop_labels=loop_labels,
        )

        etalon_search_algo = greedy_decode.GreedyRNNTInfer(
            model.decoder,
            model.joint,
            blank_index=model.joint.num_classes_with_blank - 1,
            max_symbols_per_step=5,
            preserve_alignments=True,
        )

        enc_out = encoded
        enc_len = encoded_len

        with torch.no_grad():
            hyps: list[rnnt_utils.Hypothesis] = search_algo(encoder_output=enc_out, encoded_lengths=enc_len)[0]
            hyp = decode_text_from_greedy_hypotheses(hyps, model.decoding)[0]
            etalon_hyps: list[rnnt_utils.Hypothesis] = etalon_search_algo(
                encoder_output=enc_out, encoded_lengths=enc_len
            )[0]
            etalon_hyp = decode_text_from_greedy_hypotheses(etalon_hyps, model.decoding)[0]

            assert hyp.alignments is not None
            assert etalon_hyp.alignments is not None

            assert hyp.text == etalon_hyp.text
            assert len(hyp.alignments) == len(etalon_hyp.alignments)

            for t in range(len(hyp.alignments)):
                t_u = []
                for u in range(len(hyp.alignments[t])):
                    logp, label = hyp.alignments[t][u]
                    assert torch.is_tensor(logp)
                    assert torch.is_tensor(label)
                    etalon_logp, etalon_label = etalon_hyp.alignments[t][u]
                    assert label == etalon_label
                    assert torch.allclose(logp, etalon_logp, atol=1e-4, rtol=1e-4)

                    t_u.append(int(label))

    @pytest.mark.skipif(
        not NUMBA_RNNT_LOSS_AVAILABLE,
        reason='RNNTLoss has not been compiled with appropriate numba version.',
    )
    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "beam_config",
        [
            {"search_type": "greedy"},
            {
                "search_type": "default",
                "beam_size": 2,
            },
            {
                "search_type": "alsd",
                "alsd_max_target_len": 0.5,
                "beam_size": 2,
            },
            {
                "search_type": "tsd",
                "tsd_max_sym_exp_per_step": 3,
                "beam_size": 2,
            },
            {"search_type": "maes", "maes_num_steps": 2, "maes_expansion_beta": 2, "beam_size": 2},
            {"search_type": "maes", "maes_num_steps": 3, "maes_expansion_beta": 1, "beam_size": 2},
        ],
    )
    def test_rnnt_beam_decoding_preserve_alignments(self, test_data_dir, beam_config):
        beam_size = beam_config.pop("beam_size", 1)
        model, encoded, encoded_len = get_model_encoder_output(test_data_dir, 'stt_en_conformer_transducer_small')
        beam = rnnt_beam_decoding.BeamRNNTInfer(
            model.decoder,
            model.joint,
            beam_size=beam_size,
            return_best_hypothesis=False,
            preserve_alignments=True,
            **beam_config,
        )

        enc_out = encoded
        enc_len = encoded_len
        blank_id = torch.tensor(model.joint.num_classes_with_blank - 1, dtype=torch.int32)

        with torch.no_grad():
            hyps = beam(encoder_output=enc_out, encoded_lengths=enc_len)[0]  # type: rnnt_utils.Hypothesis
            hyp, all_hyps = decode_text_from_nbest_hypotheses(hyps, model.decoding)
            hyp = hyp[0]  # best hypothesis
            all_hyps = all_hyps[0]

            assert hyp.alignments is not None

            if beam_config['search_type'] == 'alsd':
                assert len(all_hyps) <= int(beam_config['alsd_max_target_len'] * float(enc_len[0]))

            print("Beam search algorithm :", beam_config['search_type'])
            # Use the following commented print statements to check
            # the alignment of other algorithms compared to the default
            for idx, hyp_ in enumerate(all_hyps):  # type: (int, rnnt_utils.Hypothesis)
                print("Hyp index", idx + 1, "text :", hyp_.text)

                # Alignment length (T) must match audio length (T)
                # NOTE: increase length threshold to two to prevent intermittent failures when a word is split into subwords
                assert abs(len(hyp_.alignments) - enc_len[0]) <= 2  # 1

                for t in range(len(hyp_.alignments)):
                    t_u = []
                    for u in range(len(hyp_.alignments[t])):
                        logp, label = hyp_.alignments[t][u]
                        assert torch.is_tensor(logp)
                        assert torch.is_tensor(label)

                        t_u.append(int(label))

                    # Blank token must be the last token in the current
                    if len(t_u) > 1:
                        assert t_u[-1] == blank_id

                        # No blank token should be present in the current timestamp other than at the end
                        for token in t_u[:-1]:
                            assert token != blank_id

                    print(f"Tokens at timestamp {t} = {t_u}")
                print()

                assert len(hyp_.timestamp) > 0
                print("Timesteps", hyp_.timestamp)
                print()

    @pytest.mark.skipif(
        not NUMBA_RNNT_LOSS_AVAILABLE,
        reason='RNNTLoss has not been compiled with appropriate numba version.',
    )
    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "model_name, decoding_strategy",
        [
            ("stt_en_conformer_transducer_small", "greedy"),
            ("stt_en_conformer_transducer_small", "greedy_batch"),
            ("stt_en_conformer_transducer_small", "beam"),
            # ("stt_en_conformer_transducer_small", "tsd"),
            ("stt_en_conformer_transducer_small", "alsd"),
            ("nvidia/parakeet-tdt_ctc-110m", "greedy"),
            ("nvidia/parakeet-tdt_ctc-110m", "greedy_batch"),
        ],
    )
    def test_subword_decoding_compute_timestamps(self, test_data_dir, decoding_strategy, model_name):

        model, encoded, encoded_len = get_model_encoder_output(test_data_dir, model_name)

        cfg = DictConfig(model.cfg.decoding)
        cfg['strategy'] = decoding_strategy
        cfg['preserve_alignments'] = True
        cfg['compute_timestamps'] = True

        decoding = RNNTBPEDecoding(
            decoding_cfg=cfg, decoder=model.decoder, joint=model.joint, tokenizer=model.tokenizer
        )

        hyps = decoding.rnnt_decoder_predictions_tensor(encoded, encoded_len, return_hypotheses=True)
        if isinstance(hyps[0], list):
            BaseTimestampsTest.check_subword_timestamps(hyps[0][0], decoding)
        else:
            BaseTimestampsTest.check_subword_timestamps(hyps[0], decoding)

    @pytest.mark.skipif(
        not NUMBA_RNNT_LOSS_AVAILABLE,
        reason='RNNTLoss has not been compiled with appropriate numba version.',
    )
    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "model_name, decoding_strategy",
        [
            ("stt_en_conformer_transducer_small", "greedy"),
            ("stt_en_conformer_transducer_small", "greedy_batch"),
            ("stt_en_conformer_transducer_small", "beam"),
            # ("stt_en_conformer_transducer_small", "tsd"),
            ("stt_en_conformer_transducer_small", "alsd"),
            ("nvidia/parakeet-tdt_ctc-110m", "greedy"),
            ("nvidia/parakeet-tdt_ctc-110m", "greedy_batch"),
        ],
    )
    def test_char_decoding_compute_timestamps(self, test_data_dir, decoding_strategy, model_name):

        model, encoded, encoded_len = get_model_encoder_output(test_data_dir, model_name)

        cfg = DictConfig(model.cfg.decoding)
        cfg['strategy'] = decoding_strategy
        cfg['preserve_alignments'] = True
        cfg['compute_timestamps'] = True

        vocab = [t[0] for t in model.tokenizer.vocab]

        decoding = RNNTDecoding(decoding_cfg=cfg, decoder=model.decoder, joint=model.joint, vocabulary=vocab)

        hyps = decoding.rnnt_decoder_predictions_tensor(encoded, encoded_len, return_hypotheses=True)

        if isinstance(hyps[0], list):
            BaseTimestampsTest.check_char_timestamps(hyps[0][0], decoding)
        else:
            BaseTimestampsTest.check_char_timestamps(hyps[0], decoding)

    @pytest.mark.skipif(
        not NUMBA_RNNT_LOSS_AVAILABLE,
        reason='RNNTLoss has not been compiled with appropriate numba version.',
    )
    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize("use_cuda_graph_decoder", [True, False])
    @pytest.mark.parametrize("use_lm", [True, False])
    @pytest.mark.parametrize("use_boosting_tree", [True, False])
    @pytest.mark.parametrize("enable_per_stream_biasing", [True, False])
    def test_tdt_greedy_decoding(
        self,
        test_data_dir,
        use_cuda_graph_decoder: bool,
        use_lm: bool,
        use_boosting_tree: bool,
        enable_per_stream_biasing: bool,
    ):
        kenlm_model_path = Path(test_data_dir) / "asr/kenlm_ngram_lm/parakeet-tdt_ctc-110m-libri-1024.kenlm.tmp.arpa"
        boosting_tree = BoostingTreeModelConfig(key_phrases_list=["hello", "nvidia"]) if use_boosting_tree else None
        check_tdt_greedy_decoding(
            test_data_dir,
            use_cuda_graph_decoder=use_cuda_graph_decoder,
            lm_path=kenlm_model_path if use_lm else None,
            boosting_tree=boosting_tree,
            enable_per_stream_biasing=enable_per_stream_biasing,
        )

    @pytest.mark.skipif(
        not NUMBA_RNNT_LOSS_AVAILABLE,
        reason='RNNTLoss has not been compiled with appropriate numba version.',
    )
    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "beam_config",
        [
            {
                "search_type": "default",
                "beam_size": 2,
            },
            {"search_type": "maes", "maes_num_steps": 2, "maes_expansion_beta": 2, "beam_size": 2},
            {"search_type": "maes", "maes_num_steps": 2, "maes_expansion_beta": 1, "beam_size": 4},
        ],
    )
    def test_tdt_beam_decoding(self, test_data_dir, beam_config):
        check_beam_decoding(test_data_dir, beam_config)

    @pytest.mark.skipif(
        not NUMBA_RNNT_LOSS_AVAILABLE,
        reason='RNNTLoss has not been compiled with appropriate numba version.',
    )
    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "beam_config",
        [
            {
                "search_type": "maes",
                "maes_num_steps": 2,
                "maes_expansion_beta": 1,
                "beam_size": 4,
                "ngram_lm_alpha": 0.3,
            },
        ],
    )
    def test_tdt_beam_decoding_with_kenlm(self, test_data_dir, beam_config):
        # skipping if kenlm is not installed
        pytest.importorskip("kenlm", reason="Skipping test because 'kenlm' is not installed.")

        kenlm_model_path = os.path.join(
            test_data_dir, "asr", "kenlm_ngram_lm", "parakeet-tdt_ctc-110m-libri-1024.kenlm.tmp.arpa"
        )
        beam_config["ngram_lm_model"] = kenlm_model_path
        check_beam_decoding(test_data_dir, beam_config)


class TestRNNTTimestamps(BaseTimestampsTest):
    """RNNT-specific timestamp tests that inherit from BaseTimestampsTest"""

    def _convert_offsets(self, offsets):
        result = copy.deepcopy(offsets)
        for offset in result:
            offset['char'] = [offset['char']]
        return result

    @property
    def char_offsets_chars(self):
        return self._convert_offsets(super().char_offsets_chars)

    @property
    def char_offsets_wpe(self):
        return self._convert_offsets(super().char_offsets_wpe)

    @property
    def char_offsets_bpe(self):
        return self._convert_offsets(super().char_offsets_bpe)

    @property
    def encoded_char_offsets_bpe(self):
        return self._convert_offsets(super().encoded_char_offsets_bpe)

    @cached_property
    def decoding_char(self):
        cfg = RNNTDecodingConfig()
        vocab = char_vocabulary()
        decoder = get_rnnt_decoder(vocab_size=len(vocab))
        joint = get_rnnt_joint(vocab_size=len(vocab))
        decoding = RNNTDecoding(decoding_cfg=cfg, decoder=decoder, joint=joint, vocabulary=vocab)
        return decoding

    @cached_property
    def decoding_subword_wpe(self):
        cfg = RNNTDecodingConfig()
        vocab = self.tmp_tokenizer.vocab
        decoder = get_rnnt_decoder(vocab_size=len(vocab))
        joint = get_rnnt_joint(vocab_size=len(vocab))
        decoding = RNNTBPEDecoding(decoding_cfg=cfg, decoder=decoder, joint=joint, tokenizer=self.tmp_tokenizer)
        return decoding

    @cached_property
    def decoding_subword_bpe(self):
        vocab = self.bpe_tokenizer.vocab
        cfg = RNNTDecodingConfig()
        decoder = get_rnnt_decoder(vocab_size=len(vocab))
        joint = get_rnnt_joint(vocab_size=len(vocab))
        decoding = RNNTBPEDecoding(decoding_cfg=cfg, decoder=decoder, joint=joint, tokenizer=self.bpe_tokenizer)
        return decoding

    @pytest.mark.unit
    def test_word_offsets_subword_wpe(self, tmp_tokenizer):
        self.tmp_tokenizer = tmp_tokenizer
        super().test_word_offsets_subword_wpe()

    @pytest.mark.unit
    def test_word_offsets_subword_wpe_other_delimiter(self, tmp_tokenizer):
        self.tmp_tokenizer = tmp_tokenizer
        super().test_word_offsets_subword_wpe_other_delimiter()


@pytest.mark.unit
@pytest.mark.with_downloads
def test_transcribe_timestamps_no_decoder_reinstantiation(stt_en_fastconformer_transducer_large, test_data_dir):
    """
    Test that calling transcribe with timestamps=True multiple times
    does not reinstantiate the decoder.

    Regression test for the fix that avoids calling change_decoding_strategy()
    when compute_timestamps is already set to the desired value.
    """
    model = stt_en_fastconformer_transducer_large
    audio_file = os.path.join(test_data_dir, "asr/test/an4/wav/cen3-mjwl-b.wav")

    # First call - may change decoding strategy
    _ = model.transcribe(audio_file, timestamps=True)

    # Get reference to decoding algorithm after first call
    decoding_after_first_call = model.decoding.decoding

    # Second call - should NOT reinstantiate decoder
    _ = model.transcribe(audio_file, timestamps=True)

    # Verify decoder is the same object (not reinstantiated)
    assert model.decoding.decoding is decoding_after_first_call, (
        "Decoder was reinstantiated on second transcribe call with timestamps=True. "
        "This indicates change_decoding_strategy() was called unnecessarily."
    )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Optional

import lightning.pytorch as pl
import nemo_run as run
import torch

import nemo.lightning as nl
from nemo.collections import llm
from nemo.collections.llm.gpt.data.packed_sequence import PackedSequenceSpecs
from nemo.collections.llm.peft import DoRA, LoRA
from nemo.collections.llm.recipes.log.default import tensorboard_logger
from nemo.collections.llm.recipes.optim.adam import distributed_fused_adam_with_cosine_annealing
from nemo.collections.llm.recipes.precision.mixed_precision import bf16_mixed
from nemo.lightning.pytorch.callbacks import PEFT
from nemo.utils.exp_manager import TimingCallback

if TYPE_CHECKING:
    from lightning.pytorch.loggers import TensorBoardLogger, WandbLogger

TokenizerType = Any


def default_finetune_recipe(
    model: run.Config[pl.LightningModule],
    resume_path: str,
    dir: Optional[str] = None,
    name: str = "default",
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    packed_sequence: bool = False,  # once packing recipe is well tested, change this default to true
    tokenizer: Optional[TokenizerType] = "model",
) -> run.Partial:
    """
    Create a default fine-tuning recipe for any model.

    This function sets up a template for a complete configuration for fine-tuning, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        model (run.Config[pl.LightningModule]): Configuration for a NeMo model.
        resume_path (str): Path to the Huggingface model or pretrained distributed checkpoint for resume
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the fine-tuning run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        packed_sequence (bool): Whether to use packed sequence.
        tokenizer (Optional[TokenizerType]): Tokenizer setting to be applied. Can be 'data' or 'model'
            or an instance of TokenizerSpec.

    Returns:
        run.Partial: Partial configuration for fine-tuning.

    See usages of this recipe for further details.
    """
    if packed_sequence:
        datamodule = run.Config(
            llm.SquadDataModule,
            seq_length=2048,
            global_batch_size=8,
            micro_batch_size=1,
            packed_sequence_specs=PackedSequenceSpecs(packed_sequence_size=2048),
        )
    else:
        datamodule = run.Config(llm.SquadDataModule, seq_length=2048, global_batch_size=128, micro_batch_size=1)
    recipe = run.Partial(
        llm.finetune,
        model=model,
        trainer=default_finetune_trainer(
            num_nodes=num_nodes,
            num_gpus_per_node=num_gpus_per_node,
        ),
        data=datamodule,
        log=default_finetune_log(dir=dir, name=name, tensorboard_logger=tensorboard_logger(name=name)),
        optim=distributed_fused_adam_with_cosine_annealing(max_lr=1e-4, min_lr=0, warmup_steps=50, adam_beta2=0.98),
        resume=nemo_resume(resume_path),
        tokenizer=tokenizer,
    )

    return recipe


def default_finetune_trainer(
    tensor_parallelism=1,
    pipeline_parallelism=1,
    pipeline_parallelism_type=torch.bfloat16,
    virtual_pipeline_parallelism=None,
    context_parallelism=1,
    sequence_parallelism=False,
    num_nodes=1,
    num_gpus_per_node=8,
    max_steps=1000,
    limit_test_batches=None,
    limit_val_batches=None,
    val_check_interval=30,
):
    """
    Create a default fine-tuning trainer for any model.

    This function sets up a template for strategy and trainer.

    Args:
        See docstrings of MegatronStrategy and Trainer.

    Returns:
        run.Config: Config for a finetuning trainer.

    See usages of this in recipes for further details.
    """
    strategy = run.Config(
        nl.MegatronStrategy,
        tensor_model_parallel_size=tensor_parallelism,
        pipeline_model_parallel_size=pipeline_parallelism,
        pipeline_dtype=pipeline_parallelism_type,
        virtual_pipeline_model_parallel_size=virtual_pipeline_parallelism,
        context_parallel_size=context_parallelism,
        sequence_parallel=sequence_parallelism,
        gradient_as_bucket_view=True,
        ckpt_load_strictness="log_all",
    )

    trainer = run.Config(
        nl.Trainer,
        accelerator="gpu",
        accumulate_grad_batches=1,
        devices=num_gpus_per_node,
        limit_test_batches=limit_test_batches,
        limit_val_batches=limit_val_batches,
        log_every_n_steps=1,
        max_steps=max_steps,
        num_nodes=num_nodes,
        plugins=bf16_mixed(),
        strategy=strategy,
        use_distributed_sampler=False,
        val_check_interval=val_check_interval,
        callbacks=[run.Config(TimingCallback)],
    )

    return trainer


def default_finetune_log(
    dir: Optional[str] = None,
    name: str = "default",
    tensorboard_logger: Optional[run.Config['TensorBoardLogger']] = None,
    wandb_logger: Optional[run.Config['WandbLogger']] = None,
) -> run.Config[nl.NeMoLogger]:
    """
    Create a default fine-tuning logger for any model.

    This function sets up a template for ModelCheckpoint and NeMoLogger.

    Args:
        See docstrings of ModelCheckpoint and NeMoLogger.

    Returns:
        run.Config: Config for a finetuning NeMoLogger.

    See usages of this in recipes for further details.
    """

    ckpt = run.Config(
        nl.ModelCheckpoint,
        save_last="link",
        save_top_k=2,
        every_n_train_steps=50,
        filename="{model_name}--{val_loss:.2f}-{step}-{consumed_samples}",
    )

    return run.Config(
        nl.NeMoLogger,
        ckpt=ckpt,
        name=name,
        tensorboard=tensorboard_logger,
        wandb=wandb_logger,
        log_dir=dir,
    )


def nemo_resume(model_id: str) -> run.Config[nl.AutoResume]:
    """
    Configure automatic resumption from a NeMo checkpoint converted from Huggingface for
    https://huggingface.co/{model_id}.

    This NeMo checkpoint should be converted from Huggingface beforehand, using nemo.collections.llm.import_ckpt.
    When converting the checkpoint, the NeMo checkpoint will be saved in NEMO_HOME (set to ~/.cache/nemo by default).

    This function sets up the configuration to resume training from path nemo://{model_id}.
    This translates to the full path {NEMO_HOME}/models/{model_id}.

    Args:
        model_id (str): Path to the Huggingface model or pretrained distributed checkpoint for resume

    Returns:
        run.Config[nl.AutoResume]: Configuration for resuming from NeMo checkpoint.
    """
    return run.Config(
        nl.AutoResume,
        restore_config=run.Config(nl.RestoreConfig, path=f"nemo://{model_id}"),
    )


@run.cli.factory(name='lora')
def lora() -> run.Config[PEFT]:
    """
    Factory function to create a LoRA configuration.

    Returns:
        run.Config[PEFT]: Configuration for the LoRA class.

    Examples:
        CLI usage:
            $ nemo llm finetune -f llama3_8b peft=lora

        Python API usage:
            >>> lora_config = lora()
            >>> print(lora_config)
    """
    return run.Config(LoRA)


@run.cli.factory(name='dora')
def dora() -> run.Config[PEFT]:
    """
    Factory function to create a DoRA configuration.

    Returns:
        run.Config[PEFT]: Configuration for the DoRA class.

    Examples:
        CLI usage:
            $ nemo llm finetune -f llama3_8b peft=dora

        Python API usage:
            >>> dora_config = dora()
            >>> print(dora_config)
    """
    return run.Config(DoRA)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import os

import tempfile

import lightning.pytorch as pl
import numpy as np
import pytest
import soundfile as sf
import torch
from omegaconf import DictConfig, ListConfig

from nemo.collections.asr.data import audio_to_label
from nemo.collections.asr.models import EncDecClassificationModel, EncDecFrameClassificationModel, configs
from nemo.utils.config_utils import assert_dataclass_signature_match


@pytest.fixture()
def speech_classification_model():
    preprocessor = {'cls': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor', 'params': dict({})}
    encoder = {
        'cls': 'nemo.collections.asr.modules.ConvASREncoder',
        'params': {
            'feat_in': 64,
            'activation': 'relu',
            'conv_mask': True,
            'jasper': [
                {
                    'filters': 32,
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                }
            ],
        },
    }

    decoder = {
        'cls': 'nemo.collections.asr.modules.ConvASRDecoderClassification',
        'params': {
            'feat_in': 32,
            'num_classes': 30,
        },
    }

    modelConfig = DictConfig(
        {
            'preprocessor': DictConfig(preprocessor),
            'encoder': DictConfig(encoder),
            'decoder': DictConfig(decoder),
            'labels': ListConfig(["dummy_cls_{}".format(i + 1) for i in range(30)]),
        }
    )
    model = EncDecClassificationModel(cfg=modelConfig)
    return model


@pytest.fixture()
def frame_classification_model():
    preprocessor = {'cls': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor', 'params': dict({})}
    encoder = {
        'cls': 'nemo.collections.asr.modules.ConvASREncoder',
        'params': {
            'feat_in': 64,
            'activation': 'relu',
            'conv_mask': True,
            'jasper': [
                {
                    'filters': 32,
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                }
            ],
        },
    }

    decoder = {
        'cls': 'nemo.collections.common.parts.MultiLayerPerceptron',
        'params': {
            'hidden_size': 32,
            'num_classes': 5,
        },
    }

    optim = {
        'name': 'sgd',
        'lr': 0.01,
        'weight_decay': 0.001,
        'momentum': 0.9,
    }

    modelConfig = DictConfig(
        {
            'preprocessor': DictConfig(preprocessor),
            'encoder': DictConfig(encoder),
            'decoder': DictConfig(decoder),
            'optim': DictConfig(optim),
            'labels': ListConfig(["0", "1"]),
        }
    )
    model = EncDecFrameClassificationModel(cfg=modelConfig)
    return model


class TestEncDecClassificationModel:
    @pytest.mark.unit
    def test_constructor(self, speech_classification_model):
        asr_model = speech_classification_model.train()

        conv_cnt = (64 * 32 * 1 + 32) + (64 * 1 * 1 + 32)  # separable kernel + bias + pointwise kernel + bias
        bn_cnt = (4 * 32) * 2  # 2 * moving averages
        dec_cnt = 32 * 30 + 30  # fc + bias

        param_count = conv_cnt + bn_cnt + dec_cnt
        assert asr_model.num_weights == param_count

        # Check to/from config_dict:
        confdict = asr_model.to_config_dict()
        instance2 = EncDecClassificationModel.from_config_dict(confdict)

        assert isinstance(instance2, EncDecClassificationModel)

    @pytest.mark.unit
    def test_forward(self, speech_classification_model):
        asr_model = speech_classification_model.eval()

        asr_model.preprocessor.featurizer.dither = 0.0
        asr_model.preprocessor.featurizer.pad_to = 0

        input_signal = torch.randn(size=(4, 512))
        length = torch.randint(low=321, high=500, size=[4])

        with torch.no_grad():
            # batch size 1
            logprobs_instance = []
            for i in range(input_signal.size(0)):
                logprobs_ins = asr_model.forward(
                    input_signal=input_signal[i : i + 1], input_signal_length=length[i : i + 1]
                )
                logprobs_instance.append(logprobs_ins)
            logprobs_instance = torch.cat(logprobs_instance, 0)

            # batch size 4
            logprobs_batch = asr_model.forward(input_signal=input_signal, input_signal_length=length)

        assert logprobs_instance.shape == logprobs_batch.shape
        diff = torch.mean(torch.abs(logprobs_instance - logprobs_batch))
        assert diff <= 1e-6
        diff = torch.max(torch.abs(logprobs_instance - logprobs_batch))
        assert diff <= 1e-6

    @pytest.mark.unit
    def test_vocab_change(self, speech_classification_model):
        asr_model = speech_classification_model.train()

        old_labels = copy.deepcopy(asr_model._cfg.labels)
        nw1 = asr_model.num_weights
        asr_model.change_labels(new_labels=old_labels)
        # No change
        assert nw1 == asr_model.num_weights
        new_labels = copy.deepcopy(old_labels)
        new_labels.append('dummy_cls_31')
        new_labels.append('dummy_cls_32')
        new_labels.append('dummy_cls_33')
        asr_model.change_labels(new_labels=new_labels)
        # fully connected + bias
        assert asr_model.num_weights == nw1 + 3 * (asr_model.decoder._feat_in + 1)

    @pytest.mark.unit
    def test_transcription(self, speech_classification_model, test_data_dir):
        # Ground truth labels = ["yes", "no"]
        audio_filenames = ['an22-flrp-b.wav', 'an90-fbbh-b.wav']
        audio_paths = [os.path.join(test_data_dir, "asr", "train", "an4", "wav", fp) for fp in audio_filenames]

        model = speech_classification_model.eval()

        # Test Top 1 classification transcription
        results = model.transcribe(audio_paths, batch_size=2)
        assert len(results) == 2
        assert results[0].shape == torch.Size([1])

        # Test Top 5 classification transcription
        model._accuracy.top_k = [5]  # set top k to 5 for accuracy calculation
        results = model.transcribe(audio_paths, batch_size=2)
        assert len(results) == 2
        assert results[0].shape == torch.Size([5])

        # Test Top 1 and Top 5 classification transcription
        model._accuracy.top_k = [1, 5]
        results = model.transcribe(audio_paths, batch_size=2)
        assert len(results) == 2
        assert results[0].shape == torch.Size([2, 1])
        assert results[1].shape == torch.Size([2, 5])
        assert model._accuracy.top_k == [1, 5]

        # Test log probs extraction
        model._accuracy.top_k = [1]
        results = model.transcribe(audio_paths, batch_size=2, logprobs=True)
        assert len(results) == 2
        assert results[0].shape == torch.Size([len(model.cfg.labels)])

        # Test log probs extraction remains same for any top_k
        model._accuracy.top_k = [5]
        results = model.transcribe(audio_paths, batch_size=2, logprobs=True)
        assert len(results) == 2
        assert results[0].shape == torch.Size([len(model.cfg.labels)])

    @pytest.mark.unit
    def test_EncDecClassificationDatasetConfig_for_AudioToSpeechLabelDataset(self):
        # ignore some additional arguments as dataclass is generic
        IGNORE_ARGS = [
            'is_tarred',
            'num_workers',
            'batch_size',
            'tarred_audio_filepaths',
            'shuffle',
            'pin_memory',
            'drop_last',
            'tarred_shard_strategy',
            'shuffle_n',
            # `featurizer` is supplied at runtime
            'featurizer',
            # additional ignored arguments
            'vad_stream',
            'int_values',
            'sample_rate',
            'normalize_audio',
            'augmentor',
            'bucketing_batch_size',
            'bucketing_strategy',
            'bucketing_weights',
        ]

        REMAP_ARGS = {'trim_silence': 'trim'}

        result = assert_dataclass_signature_match(
            audio_to_label.AudioToSpeechLabelDataset,
            configs.EncDecClassificationDatasetConfig,
            ignore_args=IGNORE_ARGS,
            remap_args=REMAP_ARGS,
        )
        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None


class TestEncDecFrameClassificationModel(TestEncDecClassificationModel):
    @pytest.mark.parametrize(["logits_len", "labels_len"], [(20, 10), (21, 10), (19, 10), (20, 9), (20, 11)])
    @pytest.mark.unit
    def test_reshape_labels(self, frame_classification_model, logits_len, labels_len):
        model = frame_classification_model.eval()

        logits = torch.ones(4, logits_len, 2)
        labels = torch.ones(4, labels_len)
        logits_len = torch.tensor([6, 7, 8, 9])
        labels_len = torch.tensor([5, 6, 7, 8])
        labels_new, labels_len_new = model.reshape_labels(
            logits=logits, labels=labels, logits_len=logits_len, labels_len=labels_len
        )
        assert labels_new.size(1) == logits.size(1)
        assert torch.equal(labels_len_new, torch.tensor([6, 7, 8, 9]))

    @pytest.mark.unit
    def test_EncDecClassificationDatasetConfig_for_AudioToMultiSpeechLabelDataset(self):
        # ignore some additional arguments as dataclass is generic
        IGNORE_ARGS = [
            'is_tarred',
            'num_workers',
            'batch_size',
            'tarred_audio_filepaths',
            'shuffle',
            'pin_memory',
            'drop_last',
            'tarred_shard_strategy',
            'shuffle_n',
            # `featurizer` is supplied at runtime
            'featurizer',
            # additional ignored arguments
            'vad_stream',
            'int_values',
            'sample_rate',
            'normalize_audio',
            'augmentor',
            'bucketing_batch_size',
            'bucketing_strategy',
            'bucketing_weights',
            'delimiter',
            'normalize_audio_db',
            'normalize_audio_db_target',
            'window_length_in_sec',
            'shift_length_in_sec',
        ]

        REMAP_ARGS = {'trim_silence': 'trim'}

        result = assert_dataclass_signature_match(
            audio_to_label.AudioToMultiLabelDataset,
            configs.EncDecClassificationDatasetConfig,
            ignore_args=IGNORE_ARGS,
            remap_args=REMAP_ARGS,
        )
        signatures_match, cls_subset, dataclass_subset = result

        assert signatures_match
        assert cls_subset is None
        assert dataclass_subset is None

    @pytest.mark.unit
    def test_frame_classification_model(self, frame_classification_model: EncDecFrameClassificationModel):
        with tempfile.TemporaryDirectory() as temp_dir:
            # generate random audio
            audio = np.random.randn(16000 * 1)
            # save the audio
            audio_path = os.path.join(temp_dir, "audio.wav")
            sf.write(audio_path, audio, 16000)

            dummy_labels = "0 0 0 0 1 1 1 1 0 0 0 0"

            dummy_sample = {
                "audio_filepath": audio_path,
                "offset": 0.0,
                "duration": 1.0,
                "label": dummy_labels,
            }

            # create a manifest file
            manifest_path = os.path.join(temp_dir, "dummy_manifest.json")
            with open(manifest_path, "w") as f:
                for i in range(4):
                    f.write(json.dumps(dummy_sample) + "\n")

            dataloader_cfg = {
                "batch_size": 2,
                "manifest_filepath": manifest_path,
                "sample_rate": 16000,
                "num_workers": 0,
                "shuffle": False,
                "labels": ["0", "1"],
            }

            trainer_cfg = {
                "max_epochs": 1,
                "devices": 1,
                "accelerator": "auto",
            }

            optim = {
                'name': 'sgd',
                'lr': 0.01,
                'weight_decay': 0.001,
                'momentum': 0.9,
            }

            trainer = pl.Trainer(**trainer_cfg)
            frame_classification_model.set_trainer(trainer)
            frame_classification_model.setup_optimization(DictConfig(optim))
            frame_classification_model.setup_training_data(dataloader_cfg)
            frame_classification_model.setup_validation_data(dataloader_cfg)

            trainer.fit(frame_classification_model)



# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# Copyright (c) 2024 Arc Institute. All rights reserved.
# Copyright (c) 2024 Michael Poli. All rights reserved.
# Copyright (c) 2024 Stanford University. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import asdict
from pathlib import Path
from typing import Literal, Optional

import lightning.pytorch as pl
import nemo_run as run
import torch
from lightning.pytorch.callbacks.callback import Callback
from megatron.core.distributed import DistributedDataParallelConfig

from nemo import lightning as nl
from nemo.collections import llm
from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec
from nemo.collections.common.tokenizers.tokenizer_utils import get_nmt_tokenizer
from nemo.collections.llm.api import finetune, pretrain
from nemo.collections.llm.gpt.data import PreTrainingDataModule
from nemo.collections.llm.gpt.data.megatron.hyena import Evo2Dataset, parse_dataset_config
from nemo.collections.llm.gpt.data.mock import MockDataModule
from nemo.collections.llm.recipes.log.default import default_log, tensorboard_logger, wandb_logger
from nemo.collections.llm.recipes.optim.adam import distributed_fused_adam_with_cosine_annealing
from nemo.collections.llm.recipes.precision.mixed_precision import bf16_mixed, bf16_with_fp8_mixed
from nemo.collections.llm.recipes.tp_overlap_configs.userbuffers import userbuffers_bf16_h100_h8192_tp4_mbs1_seqlen8192
from nemo.lightning.pytorch import callbacks as nl_callbacks
from nemo.lightning.pytorch.callbacks.flops_callback import FLOPsMeasurementCallback
from nemo.lightning.pytorch.callbacks.megatron_comm_overlap import MegatronCommOverlapCallback
from nemo.utils.exp_manager import TimingCallback

NAME = "hyena_test"


def tokenizer_recipe():
    """
    Creates and returns a configuration for initializing a tokenizer.

    The configuration is set up to use the `get_nmt_tokenizer` function with
    the specified library type as 'byte-level'.

    Returns:
        run.Config: A configuration object for the tokenizer setup.
    """
    return run.Config(
        get_nmt_tokenizer,
        library='byte-level',
    )


def model_recipe(model_size: str, tp_comm_overlap: bool, seq_length: int) -> run.Config[pl.LightningModule]:
    """
    Factory function to create a striped hyena model

    Returns:
        run.Config[pl.LightningModule]: Configuration for the striped hyena model.

    Examples:
        CLI usage:
            $ nemo llm pretrain model=hyena_test ...

        Python API usage:
            >>> model_config = model()
            >>> print(model_config)
    """
    if model_size == 'test':
        cfg_cls = llm.HyenaTestConfig
    elif model_size == '1b':
        cfg_cls = llm.Hyena1bConfig
    elif model_size == '7b':
        cfg_cls = llm.Hyena7bConfig
    elif model_size == 'nv-7b':
        cfg_cls = llm.HyenaNV7bConfig
    elif model_size == '40b':
        cfg_cls = llm.Hyena40bConfig
    elif model_size == 'nv-40b':
        cfg_cls = llm.HyenaNV40bConfig
    elif model_size == '7b_arc_longcontext':
        cfg_cls = llm.Hyena7bARCLongContextConfig
    elif model_size == '40b_arc_longcontext':
        cfg_cls = llm.Hyena40bARCLongContextConfig
    else:
        raise NotImplementedError(f"Unsupported model size: {model_size}")

    return run.Config(
        llm.HyenaModel,
        config=run.Config(
            cfg_cls,
            seq_length=seq_length,
            tp_comm_overlap=tp_comm_overlap,
        ),
        tokenizer=tokenizer(),
    )


def trainer_recipe(
    tensor_parallelism: int = 8,
    pipeline_parallelism: int = 1,
    pipeline_parallelism_type: Optional[torch.dtype] = None,
    virtual_pipeline_parallelism: Optional[int] = None,
    context_parallelism: int = 1,
    sequence_parallelism: bool = False,
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    max_steps: int = 100,
    val_check_interval: int = 100,
    limit_test_batches: int = 50,
    limit_val_batches: int = 32,
    log_every_n_steps: int = 10,
    callbacks: Optional[list[run.Config[Callback]]] = None,
    fp8: bool = False,
    grad_reduce_in_fp32: bool = True,
    align_param_gather: bool = True,
    no_aligned_megatron_ddp: bool = False,
    ckpt_async_save: bool = True,
    save_ckpt_format: Literal['torch_dist', 'zarr'] = 'torch_dist',
) -> run.Config[nl.Trainer]:
    """
    Configure the NeMo Lightning Trainer for Striped Hyena model.

    This function sets up the distributed training strategy and other training parameters.

    Args:
        tensor_parallelism (int): Number of tensor replicas for vector parallelism.
        pipeline_parallelism (int): Number of pipeline segments for model pipeline parallelism.
        pipeline_parallelism_type ([type]): Type of pipeline parallelism to apply. Support 'interleaved','split'.
        virtual_pipelien_parallelism (int): Number of virtual pipeline stages for interleaving smaller sub-microbatches
            to reduce the computational graph bubbles caused by pipeline parallel (if pipeline parallel>1 is in use)
        context_parallelism (int): Number of context parallel blocks for processing sub-attention matrices in a block
            parallel fassion.
        sequence_parallelism (bool): Whether to use the sequence_parallelism improvement
            on the base tensor parallelism.
            This will allow for more layers to be parallelized when using tensor parallelism.
        num_nodes (int): Number of nodes for the distributed setting.
        num_gpus_per_node (int): Number of GPUs per node.
        max_steps (int): Maximum number of training steps before training terminates.
        val_check_interval (int): Interval between val check runs.
        limit_test_batches (int): Maximum number of batches over which to run test check runs on the dev set data.
        limit_val_batches (int): Maximum number of batches over which to run val check runs on the dev set data.
        log_every_n_steps (int): Log to the endpoint every n steps.
        callbacks (list[run.Config[Callback]]): A list of nemo/lightning callbacks to execute during training.
        fp8 (bool): Whether to use fp8 precision for computations.
        grad_reduce_in_fp32 (bool): Boolean indicating whether to reduce the gradient weight in the FP32 format rather
            than the default bf16.
        align_param_gather (bool): Optimization for faster train step timing potentially through aligning parameter
            gather operations.
        no_aligned_megatron_ddp (bool): Skip the aligned megatron DDP optimizations.
        ckpt_async_save (bool): Bool indicating whether to use asynchronous checkpoint saving.
        save_ckpt_format (Literal['torch_dist', 'zarr']): The checkpoint save method. The default torch_dist may result
            in larger checkpoints currently, but is the preferred option in the long run.
    Returns:
        run.Config[nl.Trainer]: Configuration for the NeMo Lightning Trainer.

    """
    if no_aligned_megatron_ddp:
        ddp: str | DistributedDataParallelConfig = run.Config(
            DistributedDataParallelConfig,
            check_for_nan_in_grad=True,
            grad_reduce_in_fp32=grad_reduce_in_fp32,
            align_param_gather=align_param_gather,
        )
    else:
        ddp = run.Config(
            DistributedDataParallelConfig,
            check_for_nan_in_grad=True,
            grad_reduce_in_fp32=grad_reduce_in_fp32,
            overlap_grad_reduce=True,
            overlap_param_gather=True,
            average_in_collective=True,
            align_param_gather=align_param_gather,
            use_distributed_optimizer=True,  # this should inherit from the optimizer config, but just in case...
        )

    strategy = run.Config(
        nl.MegatronStrategy,
        tensor_model_parallel_size=tensor_parallelism,
        pipeline_model_parallel_size=pipeline_parallelism,
        pipeline_dtype=pipeline_parallelism_type,
        virtual_pipeline_model_parallel_size=virtual_pipeline_parallelism,
        context_parallel_size=context_parallelism,
        sequence_parallel=sequence_parallelism,
        gradient_as_bucket_view=True,
        ckpt_async_save=ckpt_async_save,
        save_ckpt_format=save_ckpt_format,
        ckpt_parallel_load=True,
        ddp=ddp,
    )
    if fp8:
        mixed_precision_cfg = bf16_with_fp8_mixed()
        mixed_precision_cfg.fp8_amax_history_len = 16  # from Arc's training setup
    else:
        mixed_precision_cfg = bf16_mixed()

    trainer = run.Config(
        nl.Trainer,
        accelerator="gpu",
        accumulate_grad_batches=1,
        callbacks=callbacks,
        devices=num_gpus_per_node,
        max_steps=max_steps,
        num_nodes=num_nodes,
        plugins=mixed_precision_cfg,
        strategy=strategy,
        use_distributed_sampler=False,
        val_check_interval=val_check_interval,
        limit_test_batches=limit_test_batches,
        limit_val_batches=limit_val_batches,
        log_every_n_steps=log_every_n_steps,
    )

    return trainer


def blended_dataset_config_recipe(config_path: Path | str | None = None):
    """
    Creates a configuration for a blended dataset by utilizing the `run.Config` function.

    Args:
        config_path (Path | str | None, optional): The path to the dataset configuration file.
            Can be a `Path` object, a string, or `None`. Defaults to `None`.

    Returns:
        run.Config: A configuration object initialized with the dataset parsing function
        and the provided dataset configuration path.
    """
    return run.Config(
        parse_dataset_config,
        dataset_config_path=config_path,
    )


def pretrain_recipe_creater(
    dataset_config: str = None,
    global_batch_size: int = 8,
    micro_batch_size: int = 1,
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    grad_acc_batches: int = 1,
    tensor_parallel_size: int = 8,
    pipeline_model_parallel_size: int = 1,
    context_parallel_size: int = 1,
    seq_length: int = 8192,
    seed: int = 1234,
    model_size: str = '7b',
    use_megatron_comm_overlap_llama3_8k: bool = False,
    workers: int = 10,
    val_check_interval: int = 100,
    dir: str = None,
    enable_preemption: bool = True,
    align_param_gather: bool = True,
    tflops_callback: bool = None,
    gc_interval: int = 0,
    nsys_profiling: bool = False,
    max_steps: int = 100,
    nsys_start_step: int = 1,
    nsys_end_step: int = None,
    nsys_ranks: list[int] = [0],
    no_aligned_megatron_ddp: bool = False,
    grad_reduce_in_fp32: bool = False,
    fp8: bool = True,
    ckpt_async_save: bool = True,
    sequence_parallel: bool = True,
    ckpt_format: Literal["zarr", "torch_dist"] = "torch_dist",
    limit_val_batches: int = 10,
    restore_optimizer_from_ckpt: bool = False,
    resume_path: str = None,
    wandb_project: str = None,
    wandb_name: str = None,
    name: str = "default",
    fn=pretrain,
    **kwargs,
) -> run.Partial:
    """
    Create a pre-training recipe for a striped hyena model.

    This function sets up a complete configuration for pre-training, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        dataset_config (str): a string specifying the path to the dataset config file (json schema)
        global_batch_size (int | None): global training batch size. If left None this will be inferred from cluster and
            model parallelism configurations.
        micro_batch_size (int): micro batch size per active device.
        num_nodes (int): number of nodes to use
        num_gpus_per_node (int): number of num_gpus_per_node per node to use.
        grad_acc_batches (int): number of training batches the gradients will be accumulated before performing each
            optimizer update.
        tensor_parallel_size (int): Number of tensor parallel splits. Typically between 1 and 8 to keep parallelism
            within a single node.
        pipeline_model_parallel_size (int): Pipeline model parallel size, splits the model by layer.
        context_parallel_size (int): Context model parallel size, how many splits to make across the sequence dimension
            to be processed in parallel similar to the strategy described in the ring attention paper.
        seq_length (int): The desired sequence length to train this model on.
        seed (int): Random seed to use for initialization
        model_size (str): model size to load
        use_megatron_comm_overlap_llama3_8k (bool): If using TP, this controls advanced overlap communications
            which can improve performance during pretraining.
        workers (int): Number of workers to use for per-device batch creation.
        val_check_interval (int): How often the model evaluates during training.
        dir (str): Directory to save logs and checkpoints
        enable_preemption (bool): Enable preemption when training on slurm, captures timeout signals and attempts to
            save a final checkpoint.
        align_param_gather (bool): Optimization for faster train step timing potentially through aligning parameter
            gather operations.
        tflops_callback (bool): Enable tflops callbacks for reporting training speed and device utilization.
        gc_interval (int): How often to run GC operations throughout training (default is auto)
        nsys_profiling (bool): Enable nsys profiling from  NeMo repo.
        max_steps (int): Maximum number of steps the training model should take.
        nsys_start_step (int): Step for when NSYS will start collecting logs
        nsys_end_step (int): Step for when NSYS will stop collecting logs.
        nsys_ranks (list[int]): Ranks for processing nsys logs. Defaults to [0] if not specified.
        no_aligned_megatron_ddp (bool): Disables aligned megatron ddp optimizations.
        fp8 (bool): Whether to use fp8 precision for computations.
        grad_reduce_in_fp32 (bool): Boolean indicating whether to reduce the gradient weight in the FP32 format rather
            than the default bf16.
        ckpt_async_save (bool): Bool indicating whether to use asynchronous checkpoint saving.
        save_ckpt_format (Literal['torch_dist', 'zarr']): The checkpoint save method. The default torch_dist may result
            in larger checkpoints currently, but is the preferred option in the long run.
        resume_path (str): If specified starting weights will be loaded from this checkpoint rather than being
            randomly initialized.
        restore_optimizer_from_ckpt (bool): when loading checkpoint, try to load the optimizer.
        wandb_project (str): if set, logging to wandb will happen
        wandb_name (str): override default name for the wandb log.
    Returns:
        run.Partial: Partial configuration for pre-training.

    """
    model_run_cfg = model_recipe(
        model_size=model_size, seq_length=seq_length, tp_comm_overlap=use_megatron_comm_overlap_llama3_8k
    )
    if not dataset_config:
        data_run_cfg = run.Config(
            MockDataModule,
            seq_length=seq_length,
            global_batch_size=global_batch_size,
            micro_batch_size=micro_batch_size,
            tokenizer=tokenizer_recipe(),
        )
    else:
        data_run_cfg = run.Config(
            PreTrainingDataModule,
            paths=blended_dataset_config_recipe(dataset_config),
            dataset_cls=Evo2Dataset,
            seq_length=seq_length,
            micro_batch_size=micro_batch_size,
            global_batch_size=global_batch_size,
            seed=seed,
            num_workers=workers,
            tokenizer=tokenizer_recipe(),
        )

    extra_loggers = {}
    if wandb_project is not None:
        if wandb_name is None:
            wandb_name = (
                f"heyna-size-{model_size}-TP{tensor_parallel_size}-"
                f"PP{pipeline_model_parallel_size}-CP{context_parallel_size}"
                f"-GBS{global_batch_size}-MBS{micro_batch_size}"
                f"-GRFP32{grad_reduce_in_fp32}-"
                f"ALIGN{not no_aligned_megatron_ddp}"
                f"-NODES{num_nodes}-FP8{fp8}"
            )
        extra_loggers['wandb_logger'] = wandb_logger(project=wandb_project, name=wandb_name)
    if resume_path:
        restore_cfg = run.Config(
            nl.RestoreConfig,
            path=resume_path,
            load_model_state=True,
            load_optim_state=restore_optimizer_from_ckpt,
        )
    else:
        restore_cfg = None

    nemo_resume = run.Config(
        nl.AutoResume,
        resume_if_exists=True,
        resume_ignore_no_checkpoint=True,
        resume_past_end=False,
        resume_from_directory=dir,
        restore_config=restore_cfg,
    )

    callbacks: list[run.Config] = [
        run.Config(TimingCallback),
    ]
    if use_megatron_comm_overlap_llama3_8k:
        callbacks.append(
            run.Config(
                MegatronCommOverlapCallback,
                tp_comm_overlap=use_megatron_comm_overlap_llama3_8k,
                tp_comm_overlap_cfg=userbuffers_bf16_h100_h8192_tp4_mbs1_seqlen8192,
                wgrad_deferral_limit=22,  # default from NeMo
                overlap_param_gather_with_optimizer_step=False,
                align_param_gather=align_param_gather,
            )
        )

    if gc_interval > 0:
        callbacks.append(
            run.Config(
                nl_callbacks.GarbageCollectionCallback,
                gc_interval_train=gc_interval,
                gc_interval_val=gc_interval,
            )
        )
    if nsys_profiling:
        if nsys_end_step is None:
            nsys_end_step = max_steps
        callbacks.append(
            run.Config(
                nl_callbacks.NsysCallback,
                start_step=nsys_start_step,
                end_step=nsys_end_step,
                ranks=nsys_ranks,
                gen_shape=True,
            )
        )
    if enable_preemption:
        callbacks.append(run.Config(nl_callbacks.PreemptionCallback))
    if tflops_callback:
        # Add callback that logs the tera-FLOPS per second per GPU during training.
        flop_meas_callback = run.Config(
            FLOPsMeasurementCallback,
            run.Config(asdict, model_run_cfg),
            data_run_cfg,
            "hyena",
        )
        callbacks.append(flop_meas_callback)

    return run.Partial(
        fn,
        model=model_run_cfg,
        trainer=trainer_recipe(
            max_steps=max_steps,
            num_nodes=num_nodes,
            tensor_parallelism=tensor_parallel_size,
            pipeline_parallelism=pipeline_model_parallel_size,
            context_parallelism=context_parallel_size,
            sequence_parallelism=sequence_parallel,  # TODO Turn it on by default if TP is on
            num_gpus_per_node=num_gpus_per_node,
            val_check_interval=val_check_interval,
            limit_test_batches=limit_val_batches,
            limit_val_batches=limit_val_batches,
            callbacks=callbacks,
            fp8=fp8,
            grad_reduce_in_fp32=grad_reduce_in_fp32,
            align_param_gather=align_param_gather,
            no_aligned_megatron_ddp=no_aligned_megatron_ddp,
            ckpt_async_save=ckpt_async_save,
            save_ckpt_format=ckpt_format,
        ),
        data=data_run_cfg,
        log=default_log(
            dir=dir,
            name=name,
            tensorboard_logger=tensorboard_logger(name=name),
            **extra_loggers,
        ),
        optim=distributed_fused_adam_with_cosine_annealing(max_lr=3e-4, min_lr=3e-5, warmup_steps=2500),
        resume=nemo_resume,
    )


@run.cli.factory(name=NAME)
def tokenizer() -> run.Config[TokenizerSpec]:
    """
    Creates and returns a tokenizer configuration.

    Returns:
        run.Config[TokenizerSpec]: A configuration object for the tokenizer.
    """
    return tokenizer_recipe()


@run.cli.factory(name=NAME)
def model(tp_comm_overlap: bool, seq_length: int) -> run.Config[pl.LightningModule]:
    """
    Factory function to create a Hyena model configuration.

    Returns:
        run.Config[pl.LightningModule]: Configuration for a Hyena model.

    """
    return model_recipe('test', tp_comm_overlap, seq_length)


@run.cli.factory(target=pretrain, name=NAME)
def pretrain_recipe(
    *args,
    fn=pretrain,
    **kwargs,
) -> run.Partial:
    """
    Create a pre-training recipe for a Hyena model.

    This function sets up a complete configuration for pre-training, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the pre-training run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        fn (Callable): The pre-training function to use.

    Returns:
        run.Partial: Partial configuration for pre-training.

    """
    return pretrain_recipe_creater(*args, **kwargs)


@run.cli.factory(target=finetune, name=NAME)
def finetune_recipe(
    resume_path: str | None,
    *args,
    fn=finetune,
    **kwargs,
) -> run.Partial:
    """ """
    assert resume_path is not None, "resume_path None, invalid for finetune"
    return pretrain_recipe_creater(*args, resume_path=resume_path, fn=fn, **kwargs)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=line-too-long

from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Callable, Optional, Union

import torch
from megatron.core.transformer.transformer_config import TransformerConfig
from torch import nn

from nemo.collections.llm import Llama2Config7B, Llama2Config13B, LlamaConfig
from nemo.collections.llm.utils import Config
from nemo.collections.vlm.neva.model.base import NevaConfig, NevaModel
from nemo.collections.vlm.vision.base import HFCLIPVisionConfig, MultimodalProjectorConfig
from nemo.lightning import OptimizerModule, io, teardown

if TYPE_CHECKING:

    from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer
    from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec


# Note: these Llava configs are copied from the corresponding HF model. You may need to modify the parameter for
# your own needs


@dataclass
class LlavaConfig(NevaConfig):
    """Llava Model Base Config"""

    drop_vision_class_token: bool = True
    freeze_vision_model: bool = True


@dataclass
class Llava15Config7B(LlavaConfig):
    """Llava v1.5 Config 7B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Llama2Config7B())
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: HFCLIPVisionConfig(pretrained_model_name_or_path="openai/clip-vit-large-patch14-336")
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(input_size=1024, hidden_size=4096, ffn_hidden_size=4096)
    )


@dataclass
class Llava15Config13B(LlavaConfig):
    """Llava v1.5 Config 13B"""

    from transformers import PretrainedConfig

    language_transformer_config: TransformerConfig = field(default_factory=lambda: Llama2Config13B())
    vision_transformer_config: Union[TransformerConfig, PretrainedConfig] = field(
        default_factory=lambda: HFCLIPVisionConfig(pretrained_model_name_or_path="openai/clip-vit-large-patch14-336")
    )
    vision_projection_config: TransformerConfig = field(
        default_factory=lambda: MultimodalProjectorConfig(input_size=1024, hidden_size=5120, ffn_hidden_size=5120)
    )


class LlavaModel(NevaModel):
    """Llava Model NeMo Wrapper"""

    def __init__(
        self,
        config: Annotated[Optional[LlavaConfig], Config[LlavaConfig]] = None,
        optim: Optional[OptimizerModule] = None,
        tokenizer: Optional["TokenizerSpec"] = None,
        model_transform: Optional[Callable[[nn.Module], nn.Module]] = None,
    ):
        super().__init__(config or LlavaConfig(), optim=optim, tokenizer=tokenizer, model_transform=model_transform)


@io.model_importer(LlavaModel, "hf")
class HFLlavaImporter(io.ModelConnector["LlavaForConditionalGeneration", LlavaModel]):
    """Llava Model HF Importer"""

    def init(self) -> LlavaModel:
        # pylint: disable=C0115,C0116
        return LlavaModel(self.config, tokenizer=self.tokenizer)

    def apply(self, output_path: Path) -> Path:
        # pylint: disable=C0115,C0116
        from transformers import LlavaForConditionalGeneration

        source = LlavaForConditionalGeneration.from_pretrained(str(self))
        target = self.init()
        trainer = self.nemo_setup(target)
        self.convert_state(source, target)
        print(f"Converted Llava model to Nemo, saving to {output_path}")

        self.nemo_save(output_path, trainer)

        print(f"Converted Llava model saved to {output_path}")

        teardown(trainer, target)
        del trainer, target

        return output_path

    def convert_state(self, source, target, image_newline=False):
        # pylint: disable=C0115,C0116
        mapping = {
            "model.language_model.embed_tokens.weight": "language_model.embedding.word_embeddings.weight",
            "model.language_model.layers.*.self_attn.o_proj.weight": "language_model.decoder.layers.*.self_attention.linear_proj.weight",
            "model.language_model.layers.*.mlp.down_proj.weight": "language_model.decoder.layers.*.mlp.linear_fc2.weight",
            "model.language_model.layers.*.input_layernorm.weight": "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight",
            "model.language_model.layers.*.post_attention_layernorm.weight": "language_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight",
            "model.language_model.norm.weight": "language_model.decoder.final_layernorm.weight",
            "lm_head.weight": "language_model.output_layer.weight",
        }

        if "vision_projection.encoder.linear_fc1.weight" in target.module.state_dict().keys():
            mapping.update(
                {
                    "model.multi_modal_projector.linear_1.weight": "vision_projection.encoder.linear_fc1.weight",
                    "model.multi_modal_projector.linear_1.bias": "vision_projection.encoder.linear_fc1.bias",
                    "model.multi_modal_projector.linear_2.weight": "vision_projection.encoder.linear_fc2.weight",
                    "model.multi_modal_projector.linear_2.bias": "vision_projection.encoder.linear_fc2.bias",
                }
            )
        elif "vision_projection.0.weight" in target.module.state_dict().keys():
            mapping.update(
                {
                    "model.multi_modal_projector.linear_1.weight": "vision_projection.0.weight",
                    "model.multi_modal_projector.linear_1.bias": "vision_projection.0.bias",
                    "model.multi_modal_projector.linear_2.weight": "vision_projection.2.weight",
                    "model.multi_modal_projector.linear_2.bias": "vision_projection.2.bias",
                }
            )
        else:
            raise KeyError("Unable to map vision projection keys.")

        if image_newline:
            mapping.update({"model.image_newline": "image_newline"})

        if "vision_model.vision_model.embeddings.class_embedding" in target.module.state_dict().keys():
            mapping.update(
                {
                    "model.vision_tower.vision_model.**": "vision_model.vision_model.**",
                }
            )
        elif "vision_model.class_token" in target.module.state_dict().keys():
            mapping.update(
                {
                    "model.vision_tower.vision_model.embeddings.patch_embedding.weight": "vision_model.conv1.weight",
                    "model.vision_tower.vision_model.embeddings.position_embedding.weight": "vision_model.position_embeddings.weight",
                    "model.vision_tower.vision_model.encoder.layers.*.layer_norm1.weight": "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight",
                    "model.vision_tower.vision_model.encoder.layers.*.layer_norm1.bias": "vision_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_bias",
                    "model.vision_tower.vision_model.encoder.layers.*.layer_norm2.weight": "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight",
                    "model.vision_tower.vision_model.encoder.layers.*.layer_norm2.bias": "vision_model.decoder.layers.*.mlp.linear_fc1.layer_norm_bias",
                    "model.vision_tower.vision_model.encoder.layers.*.self_attn.out_proj.weight": "vision_model.decoder.layers.*.self_attention.linear_proj.weight",
                    "model.vision_tower.vision_model.encoder.layers.*.self_attn.out_proj.bias": "vision_model.decoder.layers.*.self_attention.linear_proj.bias",
                    "model.vision_tower.vision_model.encoder.layers.*.mlp.fc1.weight": "vision_model.decoder.layers.*.mlp.linear_fc1.weight",
                    "model.vision_tower.vision_model.encoder.layers.*.mlp.fc1.bias": "vision_model.decoder.layers.*.mlp.linear_fc1.bias",
                    "model.vision_tower.vision_model.encoder.layers.*.mlp.fc2.weight": "vision_model.decoder.layers.*.mlp.linear_fc2.weight",
                    "model.vision_tower.vision_model.encoder.layers.*.mlp.fc2.bias": "vision_model.decoder.layers.*.mlp.linear_fc2.bias",
                    "model.vision_tower.vision_model.pre_layrnorm.weight": "vision_model.ln_pre.weight",
                    "model.vision_tower.vision_model.pre_layrnorm.bias": "vision_model.ln_pre.bias",
                }
            )
        else:
            raise KeyError("Unable to map vision encoder keys.")
        return io.apply_transforms(
            source,
            target,
            mapping=mapping,
            transforms=[
                _import_language_qkv,
                _import_vision_qkv,
                _import_vision_qkv_bias,
                _import_linear_fc1,
            ],
        )

    @property
    def tokenizer(self) -> "AutoTokenizer":
        # pylint: disable=C0115,C0116
        from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer

        return AutoTokenizer(str(self))

    @property
    def config(self) -> LlavaConfig:
        # pylint: disable=C0115,C0116
        from transformers import LlavaConfig as HFLlavaConfig

        source = HFLlavaConfig.from_pretrained(str(self))
        text_conifg = source.text_config

        def make_vocab_size_divisible_by(vocab_size):
            # pylint: disable=C0115,C0116
            base = 128
            while vocab_size % base != 0:
                base //= 2
            return base

        language_transformer_config = LlamaConfig(
            num_layers=text_conifg.num_hidden_layers,
            hidden_size=text_conifg.hidden_size,
            ffn_hidden_size=text_conifg.intermediate_size,
            num_attention_heads=text_conifg.num_attention_heads,
            init_method_std=text_conifg.initializer_range,
            layernorm_epsilon=text_conifg.rms_norm_eps,
            num_query_groups=text_conifg.num_key_value_heads,
            rotary_base=text_conifg.rope_theta,
            gated_linear_unit=True,
            make_vocab_size_divisible_by=make_vocab_size_divisible_by(text_conifg.vocab_size),
            share_embeddings_and_output_weights=False,
        )
        vision_transformer_config = HFCLIPVisionConfig(
            pretrained_model_name_or_path="openai/clip-vit-large-patch14-336"
        )
        vision_projection_config = MultimodalProjectorConfig(input_size=1024, hidden_size=4096, ffn_hidden_size=4096)

        output = LlavaConfig(
            language_transformer_config=language_transformer_config,
            vision_transformer_config=vision_transformer_config,
            vision_projection_config=vision_projection_config,
            vision_feature_layer=source.vision_feature_layer,
        )

        return output


def import_qkv(q, k, v, head_num, num_query_groups, heads_per_group, hidden_size, head_size):
    # pylint: disable=C0115,C0116
    old_tensor_shape = q.size()
    new_q_tensor_shape = (head_num, head_size) + old_tensor_shape[1:]
    new_kv_tensor_shape = (num_query_groups, head_size) + old_tensor_shape[1:]

    q = q.view(*new_q_tensor_shape)
    k = k.view(*new_kv_tensor_shape)
    v = v.view(*new_kv_tensor_shape)

    qkv_weights_l = []
    for i in range(num_query_groups):
        qkv_weights_l.append(q[i * heads_per_group : (i + 1) * heads_per_group, :, :])
        qkv_weights_l.append(k[i : i + 1, :, :])
        qkv_weights_l.append(v[i : i + 1, :, :])
    qkv_weights = torch.cat(qkv_weights_l)
    assert qkv_weights.ndim == 3, qkv_weights.shape
    assert qkv_weights.shape[0] == (heads_per_group + 2) * num_query_groups, qkv_weights.shape
    assert qkv_weights.shape[1] == head_size, qkv_weights.shape
    assert qkv_weights.shape[2] == old_tensor_shape[1], qkv_weights.shape

    qkv_weights = qkv_weights.reshape([head_size * (head_num + 2 * num_query_groups), hidden_size])

    return qkv_weights


def export_qkv(linear_qkv, head_num, num_query_groups, heads_per_group, hidden_size, head_size):
    # pylint: disable=C0115,C0116
    qkv_total_dim = head_num + 2 * num_query_groups

    linear_qkv = linear_qkv.reshape([qkv_total_dim, head_size, -1])
    hidden_size = linear_qkv.size(-1)
    q_slice = torch.cat(
        [
            torch.arange((heads_per_group + 2) * i, (heads_per_group + 2) * i + heads_per_group)
            for i in range(num_query_groups)
        ]
    )
    k_slice = torch.arange(heads_per_group, qkv_total_dim, (heads_per_group + 2))
    v_slice = torch.arange(heads_per_group + 1, qkv_total_dim, (heads_per_group + 2))

    q_proj = linear_qkv[q_slice].reshape(-1, hidden_size).cpu()
    k_proj = linear_qkv[k_slice].reshape(-1, hidden_size).cpu()
    v_proj = linear_qkv[v_slice].reshape(-1, hidden_size).cpu()

    return q_proj, k_proj, v_proj


def export_qkv_bias(qkv_bias: torch.Tensor, head_num, num_query_groups, heads_per_group, head_size):
    """
    Split interleave-concatenated qkv bias to separate q, k, v bias

    Example: export layer linear_qkv bias to HF {q|k|v}_proj bias
    """
    qkv_total_dim = head_num + 2 * num_query_groups

    qkv_bias = qkv_bias.reshape([qkv_total_dim, head_size])
    q_slice = torch.cat(
        [
            torch.arange((heads_per_group + 2) * i, (heads_per_group + 2) * i + heads_per_group)
            for i in range(num_query_groups)
        ]
    )
    k_slice = torch.arange(heads_per_group, qkv_total_dim, (heads_per_group + 2))
    v_slice = torch.arange(heads_per_group + 1, qkv_total_dim, (heads_per_group + 2))

    q_bias = qkv_bias[q_slice].reshape(-1).cpu()
    k_bias = qkv_bias[k_slice].reshape(-1).cpu()
    v_bias = qkv_bias[v_slice].reshape(-1).cpu()

    return q_bias, k_bias, v_bias


@io.state_transform(
    source_key=(
        "model.language_model.layers.*.self_attn.q_proj.weight",
        "model.language_model.layers.*.self_attn.k_proj.weight",
        "model.language_model.layers.*.self_attn.v_proj.weight",
    ),
    target_key="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
)
def _import_language_qkv(ctx: io.TransformCTX, q, k, v):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.language_transformer_config
    return import_qkv(
        q,
        k,
        v,
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=megatron_config.hidden_size,
        head_size=megatron_config.kv_channels,
    )


@io.state_transform(
    source_key=(
        "model.vision_tower.vision_model.encoder.layers.*.self_attn.q_proj.weight",
        "model.vision_tower.vision_model.encoder.layers.*.self_attn.k_proj.weight",
        "model.vision_tower.vision_model.encoder.layers.*.self_attn.v_proj.weight",
    ),
    target_key="vision_model.decoder.layers.*.self_attention.linear_qkv.weight",
)
def _import_vision_qkv(ctx: io.TransformCTX, q, k, v):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.vision_transformer_config
    return import_qkv(
        q,
        k,
        v,
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=megatron_config.hidden_size,
        head_size=megatron_config.kv_channels,
    )


@io.state_transform(
    source_key=(
        "model.vision_tower.vision_model.encoder.layers.*.self_attn.q_proj.bias",
        "model.vision_tower.vision_model.encoder.layers.*.self_attn.k_proj.bias",
        "model.vision_tower.vision_model.encoder.layers.*.self_attn.v_proj.bias",
    ),
    target_key="vision_model.decoder.layers.*.self_attention.linear_qkv.bias",
)
def _import_vision_qkv_bias(ctx: io.TransformCTX, q_bias, k_bias, v_bias):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config.vision_transformer_config
    return import_qkv(
        q_bias.unsqueeze(-1),
        k_bias.unsqueeze(-1),
        v_bias.unsqueeze(-1),
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=1,
        head_size=megatron_config.kv_channels,
    ).squeeze(-1)


@io.state_transform(
    source_key="model.vision_tower.vision_model.embeddings.class_embedding",
    target_key="vision_model.class_token",
)
def _import_cls_token(ctx: io.TransformCTX, cls_token):
    # pylint: disable=C0115,C0116
    return cls_token.reshape(1, 1, -1)


@io.state_transform(
    source_key=(
        "model.language_model.layers.*.mlp.gate_proj.weight",
        "model.language_model.layers.*.mlp.up_proj.weight",
    ),
    target_key="language_model.decoder.layers.*.mlp.linear_fc1.weight",
)
def _import_linear_fc1(down, gate):
    # pylint: disable=C0115,C0116
    return torch.cat((down, gate), axis=0)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from typing import Any, Callable, Optional, TypeVar

import lightning.pytorch as pl
from torch import nn

from nemo.utils import logging


class ModelTransform(pl.Callback):
    """
    A PyTorch Lightning callback that applies a model transformation function at the start of fitting or validation.

    This callback is designed to apply a transformation to the model when fitting or validation begins.
    This design allows for loading the original checkpoint first and then applying the transformation,
    which is particularly useful for techniques like Parameter-Efficient Fine-Tuning (PEFT).

    The transformation function is expected to be defined on the LightningModule
    as an attribute called 'model_transform'.

    Key Features:
    - Applies transformation at the start of fit or validation, not during initialization.
    - Allows loading of original checkpoints before transformation.
    - Supports PEFT and similar techniques that modify model structure.

    Example:
        >>> class MyLightningModule(pl.LightningModule):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.model = SomeModel()
        ...         self.model_transform = lambda m: SomePEFTMethod()(m)
        ...
        >>> model = MyLightningModule()
        >>> # Load original checkpoint here if needed
        >>> model.load_state_dict(torch.load('original_checkpoint.pth'))
        >>> trainer = pl.Trainer(callbacks=[ModelTransform()])
        >>> # The model will be transformed when trainer.fit() or trainer.validate() is called
        >>> trainer.fit(model)

    Note:
        The transformation is applied only once, at the start of fitting or validation,
        whichever comes first. This ensures that the model structure is modified before
        any forward passes or parameter updates occur, but after the original weights
        have been loaded.
    """

    def __init__(self):
        super().__init__()
        self.model_transform: Optional[Callable[[nn.Module], nn.Module]] = None

    def setup(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", stage: str) -> None:
        """Setups model transform"""
        logging.info(f"Setting up ModelTransform for stage: {stage}")

        if hasattr(pl_module, 'model_transform'):
            logging.info("Found model_transform attribute on pl_module")
            self.model_transform = _call_counter(pl_module.model_transform)
            pl_module.model_transform = self.model_transform
            logging.info(f"Set model_transform to: {self.model_transform}")
        else:
            logging.info("No model_transform attribute found on pl_module")

    def on_train_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """event hook"""
        self._maybe_apply_transform(trainer)

    def on_validation_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """event hook"""
        self._maybe_apply_transform(trainer)

    def _maybe_apply_transform(self, trainer):
        """Applies transform if haven't already"""
        if self._needs_to_call:
            self.apply_transform(trainer)

    def apply_transform(self, trainer):
        """Applies a model transform e.g., PeFT"""
        self.model_transform(trainer.model)
        from lightning.pytorch.utilities import model_summary

        summary = str(model_summary.summarize(trainer.lightning_module, max_depth=1))
        summary = "\n\r".join(summary.split("\n"))
        logging.info(f"After applying model_transform:\n\n\r{summary}")

    @property
    def _needs_to_call(self) -> bool:
        """boolean var to indicate whether need to run transform based on call counter"""
        return self.model_transform and self.model_transform.__num_calls__ == 0


T = TypeVar('T', bound=Callable[..., Any])


def _call_counter(func: T) -> T:
    """
    A decorator that counts the number of times a function is called.

    This decorator wraps a function and adds a '__num_calls__' attribute to it,
    which is incremented each time the function is called.

    Args:
        func (Callable): The function to be wrapped.

    Returns:
        Callable: The wrapped function with a call counter.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.__num_calls__ += 1
        return func(*args, **kwargs)

    wrapper.__num_calls__ = 0
    return wrapper  # type: ignore



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch
from nemo.collections.common.data.fallback import FallbackDataset


class DummyDataset(torch.utils.data.Dataset):
    def __getitem__(self, item):
        return item


def test_fallback_dataset():
    dset = DummyDataset()
    assert dset[0] == 0
    assert dset[None] is None

    fallback = FallbackDataset(dset)
    assert fallback[0] == 0
    assert fallback[None] == 0
    assert fallback[1] == 1
    assert fallback[None] == 1

    fallback = FallbackDataset(dset)
    assert fallback[0] == 0
    assert fallback[None] == 0
    assert fallback[None] == 0

    fallback = FallbackDataset(dset)
    assert fallback[None] is None
    assert fallback[0] == 0
    assert fallback[None] == 0



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from unittest.mock import Mock, patch

import pytest
import torch

from nemo.collections.vlm.inference.vlm_inference_controller import (
    QwenVLTextGenerationController,
    TokenizerWrapper,
    VLMTextGenerationController,
)


class TestTokenizerWrapper:
    """Test cases for TokenizerWrapper class"""

    def setup_method(self):
        """Set up test fixtures before each test method"""
        self.mock_tokenizer = Mock()
        self.mock_tokenizer.eos_token_id = 2
        self.mock_tokenizer.vocab_size = 32000
        self.tokenizer_wrapper = TokenizerWrapper(self.mock_tokenizer)

    def test_init(self):
        """Test TokenizerWrapper initialization"""
        assert self.tokenizer_wrapper.eod == 2
        assert self.tokenizer_wrapper.vocab_size is None
        assert self.tokenizer_wrapper._tokenizer == self.mock_tokenizer

    def test_detokenize(self):
        """Test detokenize method"""
        mock_tokens = [1, 2, 3]
        expected_result = "test output"
        self.mock_tokenizer.decode.return_value = expected_result

        result = self.tokenizer_wrapper.detokenize(mock_tokens)

        self.mock_tokenizer.decode.assert_called_once_with(mock_tokens, skip_special_tokens=False)
        assert result == expected_result

    def test_tokenize(self):
        """Test tokenize method"""
        mock_prompt = "test prompt"
        expected_tokens = [1, 2, 3]
        self.mock_tokenizer.encode.return_value = expected_tokens

        result = self.tokenizer_wrapper.tokenize(mock_prompt)

        self.mock_tokenizer.encode.assert_called_once_with(mock_prompt, add_special_tokens=False)
        assert result == expected_tokens


class TestVLMTextGenerationController:
    """Test cases for VLMTextGenerationController class"""

    def setup_method(self):
        """Set up test fixtures before each test method"""
        self.mock_inference_model = Mock()
        self.mock_tokenizer = Mock()
        self.mock_image_processor = Mock()
        self.mock_image_processor.size = {'height': 224, 'width': 224}

        # Mock just the parent class __init__ method to prevent real initialization
        with patch.object(VLMTextGenerationController.__bases__[0], '__init__'):
            # Create the controller
            self.controller = VLMTextGenerationController(
                self.mock_inference_model, self.mock_tokenizer, self.mock_image_processor
            )

            # Manually set attributes that would normally be set by parent class
            self.controller.tokenizer = TokenizerWrapper(self.mock_tokenizer)
            self.controller.inference_wrapped_model = self.mock_inference_model

    def test_init(self):
        """Test VLMTextGenerationController initialization"""
        assert self.controller.image_processor == self.mock_image_processor

    def test_tokenize_prompt_with_no_image(self):
        """Test tokenize_prompt method when image is None"""
        mock_prompt = "test prompt"
        mock_tokens = [1, 2, 3]
        self.mock_tokenizer.encode.return_value = mock_tokens

        tokens, image_dict = self.controller.tokenize_prompt(mock_prompt, None)

        self.mock_tokenizer.encode.assert_called_once_with(mock_prompt, add_special_tokens=False)
        assert tokens == mock_tokens
        assert 'pixel_values' in image_dict
        assert 'aspect_ratio_ids' in image_dict
        assert 'num_tiles' in image_dict
        assert image_dict['pixel_values'].shape == (1, 4, 3, 224, 224)
        assert image_dict['aspect_ratio_ids'].shape == (1,)
        assert image_dict['num_tiles'] == [0]

    def test_tokenize_prompt_with_image(self):
        """Test tokenize_prompt method when image is provided"""
        mock_prompt = "test prompt"
        mock_image = Mock()
        mock_tokens = [1, 2, 3]
        self.mock_tokenizer.encode.return_value = mock_tokens

        # Mock image processor output
        mock_processed = {
            'pixel_values': torch.randn(1, 4, 3, 224, 224),
            'aspect_ratio_ids': torch.tensor([1], dtype=torch.long),
            'num_tiles': [2],
            'other_key': 'should_be_filtered',
        }
        self.mock_image_processor.preprocess.return_value = mock_processed

        tokens, image_dict = self.controller.tokenize_prompt(mock_prompt, mock_image)

        self.mock_tokenizer.encode.assert_called_once_with(mock_prompt, add_special_tokens=False)
        self.mock_image_processor.preprocess.assert_called_once_with(mock_image, return_tensors='pt')
        assert tokens == mock_tokens
        assert 'pixel_values' in image_dict
        assert 'aspect_ratio_ids' in image_dict
        assert 'num_tiles' in image_dict
        assert 'other_key' not in image_dict
        assert image_dict['pixel_values'].shape == (4, 3, 224, 224)

    def test_prep_inference_input(self):
        """Test prep_inference_input method"""
        mock_prompts_tokens = torch.randn(2, 10)
        mock_active_requests = OrderedDict(
            [('req1', Mock(encoder_prompt='image1')), ('req2', Mock(encoder_prompt='image2'))]
        )

        expected_result = {'processed': 'data'}
        self.mock_inference_model.prep_inference_input.return_value = expected_result

        result = self.controller.prep_inference_input(
            mock_prompts_tokens, mock_active_requests, use_attention_mask=True
        )

        self.mock_inference_model.prep_inference_input.assert_called_once_with(
            prompts_tokens=mock_prompts_tokens, image_dict=['image1', 'image2']
        )
        assert result == expected_result


class TestQwenVLTextGenerationController:
    """Test cases for QwenVLTextGenerationController class"""

    def setup_method(self):
        """Set up test fixtures before each test method"""
        self.mock_inference_model = Mock()
        self.mock_tokenizer = Mock()
        self.mock_image_processor = Mock()
        self.mock_image_processor.size = {'height': 224, 'width': 224}
        self.mock_processor = Mock()

        # Mock just the parent class __init__ method to prevent real initialization
        with patch.object(QwenVLTextGenerationController.__bases__[0], '__init__'):
            # Create the controller
            self.controller = QwenVLTextGenerationController(
                self.mock_inference_model, self.mock_tokenizer, self.mock_image_processor, self.mock_processor
            )

            # Manually set attributes that would normally be set by parent class
            self.controller.inference_wrapped_model = self.mock_inference_model

    def test_init(self):
        """Test QwenVLTextGenerationController initialization"""
        assert hasattr(self.controller, 'tokenizer')
        assert hasattr(self.controller, 'processor')
        assert self.controller.processor == self.mock_processor

    def test_qwenvl_tokenizer_detokenize(self):
        """Test QwenVLTokenizer detokenize method"""
        mock_tokens = [151652, 151655, -200, 100, 200]
        expected_result = "processed output"
        self.mock_tokenizer.decode.return_value = expected_result

        # Get the QwenVLTokenizer instance
        qwen_tokenizer = self.controller.tokenizer

        result = qwen_tokenizer.detokenize(mock_tokens)

        # Verify the result is correct
        assert result == expected_result
        # Verify the mock was called (the QwenVLTokenizer uses self.mock_tokenizer internally)
        self.mock_tokenizer.decode.assert_called_once()

    def test_tokenize_prompt(self):
        """Test QwenVLTextGenerationController tokenize_prompt method"""
        mock_prompt = "test prompt"
        mock_image = Mock()

        # Mock processor output
        mock_inputs = {
            'input_ids': torch.tensor([[1, 2, 151655, 3, 4]]),
            'pixel_values': torch.randn(1, 3, 224, 224),
            'image_grid_thw': torch.tensor([1, 1, 1]),
        }
        self.mock_processor.return_value = mock_inputs

        tokens, image_dict = self.controller.tokenize_prompt(mock_prompt, mock_image)

        self.mock_processor.assert_called_once_with(
            text=[mock_prompt], images=mock_image, padding=True, return_tensors="pt"
        )

        # Check that 151655 tokens are replaced with -200
        assert tokens == [1, 2, -200, 3, 4]
        assert 'pixel_values' in image_dict
        assert 'image_grid_thw' in image_dict
        assert image_dict['pixel_values'].shape == (1, 3, 224, 224)


class TestIntegration:
    """Integration tests for the controllers"""

    def test_vlm_controller_inheritance(self):
        """Test that VLMTextGenerationController properly inherits from parent"""
        mock_inference_model = Mock()
        mock_tokenizer = Mock()
        mock_image_processor = Mock()

        # Mock just the parent class __init__ method to prevent real initialization
        with patch.object(VLMTextGenerationController.__bases__[0], '__init__'):
            controller = VLMTextGenerationController(mock_inference_model, mock_tokenizer, mock_image_processor)

            # Manually set attributes that would normally be set by parent class
            controller.tokenizer = TokenizerWrapper(mock_tokenizer)
            controller.inference_wrapped_model = mock_inference_model

            # Verify the controller was created successfully and has expected attributes
            assert controller is not None
            assert hasattr(controller, 'image_processor')

    def test_qwenvl_controller_inheritance(self):
        """Test that QwenVLTextGenerationController properly inherits from VLMTextGenerationController"""
        mock_inference_model = Mock()
        mock_tokenizer = Mock()
        mock_image_processor = Mock()
        mock_processor = Mock()

        # Mock just the parent class __init__ method to prevent real initialization
        with patch.object(QwenVLTextGenerationController.__bases__[0], '__init__'):
            controller = QwenVLTextGenerationController(
                mock_inference_model, mock_tokenizer, mock_image_processor, mock_processor
            )

            # Manually set attributes that would normally be set by parent class
            controller.tokenizer = TokenizerWrapper(mock_tokenizer)
            controller.inference_wrapped_model = mock_inference_model

            # Verify the controller was created successfully and has expected attributes
            assert controller is not None
            assert hasattr(controller, 'processor')

    def test_tokenizer_wrapper_interface(self):
        """Test that TokenizerWrapper provides the expected interface"""
        mock_tokenizer = Mock()
        mock_tokenizer.eos_token_id = 1
        wrapper = TokenizerWrapper(mock_tokenizer)

        # Test that required attributes exist
        assert hasattr(wrapper, 'eod')
        assert hasattr(wrapper, 'vocab_size')
        assert hasattr(wrapper, '_tokenizer')
        assert hasattr(wrapper, 'detokenize')
        assert hasattr(wrapper, 'tokenize')

        # Test that methods are callable
        assert callable(wrapper.detokenize)
        assert callable(wrapper.tokenize)


if __name__ == "__main__":
    pytest.main([__file__])
